/*
 * BlurredImage.cpp
 *
 * @author  E. van Schaijk
 * @author  T. Poppe
 * @version  1.0
 * @since  July, 2016
 *
 * Part of the implementation of the image deblurring algorithm described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 *
 * This class represents the blurred image.
 *
 */

#include "BlurredImage.hpp"
#include <opencv2/core/core.hpp>
#include <opencv2/opencv.hpp>
#include <iostream>
#include <stdexcept>
#include <vector>

using namespace cv;
using namespace std;

/**
 * Creates the BlurredImage object.
 * gamma correction is set to the default value 1 and an empty imageOrginal is created.
 */
BlurredImage::BlurredImage(){
  this->imageOriginal.create(8, 8, CV_64FC1);
  this->gammaCorrection =1;
  this->setImageOriginalOneChannel(imageOriginal);
}

/**
 * Creates the BlurredImage object
 *
 * @param blurredImage     matrix blurredImage, which is converted to greyscale and set as imageOriginal.
 * @param gammaCorrection  gamma correction used to rescale the intensitie values of the pixels if required.
 * @throws domain_error  if gammaCorrection < 1 or if blurredImage < 5
 */
BlurredImage::BlurredImage(cv::Mat& blurredImage, gdouble gammaCorrection) throw (domain_error){

  if (blurredImage.rows < 5 ||blurredImage.cols < 5){
    throw domain_error("BlurredImage Constructor: Size blurredImage needs to be greater then 4");
  }
  //imageOriginal = blurredImage;
  this->setImageOriginal(blurredImage);

  // check if gamma correction factor is valid
  if (gammaCorrection <= 0){
    throw domain_error("BlurredImage; Constructor: gamma correction should not be negative or zero");
  }
  this->gammaCorrection = gammaCorrection;

  this->setImageOriginalOneChannel(imageOriginal, true);

  //this->initializeOptimalSize();
}

/**
 * Resizes the original image or the oneChannelImageOriginal depending on the oneChannel boolean flag,
 * with the factor and adjust the size of the resized image for optimal fft calculations.
 * The optimal FFT size takes into account the kernelsize.
 * The additional rows and or colums are filled via the wrap bounadry Liu method
 *
 * @param factor  		factor for rezising the original image.
 * @param kernelSize 	size of kernel, used for determining the optimal fft size.
 * @param oneChannel    if true, then the ImageOriginalOneChannel is resized, otherwise the imageOriginal
 * @throws domain_error  if factor is not between 0 en 1 or if kernel size < 3
 */
void BlurredImage::resizeForFFT(gdouble factor, gint kernelSize, bool oneChannel)throw (domain_error){
  // validate input
  if(kernelSize < 3){
	throw domain_error("BlurredImage resizeForFFT: Size kernel needs to be greater then 2");
  }
  if(factor <= 0 || factor > 1){
	throw domain_error("BlurredImage resizeForFFT: Resize factor needs to be greater then 0 and smaller or equal to 1");
  }

  image.release();

  if (oneChannel){
	image = downSmpImg(imageOriginalOneChannel, factor);
  } else {
	image = downSmpImg(imageOriginal,factor);
  }

  // set resized format of image.
  sizeRescaled = image.size();

  // calculate target size for FFT
  gint height = image.rows + kernelSize -1;
  gint widht = image.cols + kernelSize - 1;
//  if (height > 4096 || widht > 4096){
//	  throw domain_error("image is too big, please try partial deconvolution");
//  }

  Size targetSize = Size(getOptimalDFTSize(widht),getOptimalDFTSize(height));

  wrapBoundaryLiu(targetSize); // fill image up to the targetsize

}

/**
 * Adds additional rows and columns to change the image to the size given as input parameter.
 * The values in the additional rows and columns are determined via the wrap boundary lui method.
 *
 * @param size  The size to which the boundary needs to be populated.
 * @throws domain_error  if size < 5
 */
void BlurredImage::wrapBoundaryLiu(Size size)throw (domain_error){
  if (size.height < 5 || size.width < 5){
    throw domain_error("BlurredImage wrapBoundaryLiu: Size  needs to be greater then 4");
  }

  gint imgLastRow = image.rows;
  gint imgLastCol = image.cols;

  // create matrix a extending the bottom of the image of the final size
  gint deltaHeight = size.height -imgLastRow+2;
  gint deltaWidth = size.width - imgLastCol+2;
  if (deltaHeight < 4 || deltaWidth < 4){
    throw domain_error("BlurredImage-wrapBoundaryLiu: deltaWeight or deltaheight is smaller then 4");
  }

  //check image depth type and if necessary convert
  if (image.depth() != CV_64F){
	image.convertTo(image, CV_64F);
  }


  int channels = image.channels();
  Mat planes[4];
  split(image, planes);

  for (int i = 0 ; i < channels ; i++){

    // calculate the border of the lower left square a
	Mat a = Mat::zeros(deltaHeight, imgLastCol, CV_64FC1);

	planes[i].row(imgLastRow-1).copyTo(a.row(0)); // copy last row of image to the first row of matrix a
	planes[i].row(0).copyTo(a.row(deltaHeight-1));	// copy first row of image to the last row of matrix a
	for (int j=0; j< deltaHeight-2;j++){
	  a.at<double>(1+j,0) = (a.at<double>(deltaHeight-1,0)*j+a.at<double>(0,0)*(deltaHeight-3-j))/(deltaHeight-3);
	  a.at<double>(1+j,imgLastCol-1) = (a.at<double>(deltaHeight-1,imgLastCol-1)*j
				  +a.at<double>(0,imgLastCol-1)*(deltaHeight-3-j))/(deltaHeight-3);
	}
	this->solveMinLaPlacion(a);

	// calculate the border of the upper right square b
	Mat b = Mat::zeros(imgLastRow, deltaWidth, CV_64FC1);
	planes[i].col(imgLastCol-1).copyTo(b.col(0)); // copy last col of image to the first col of matrix b
	planes[i].col(0).copyTo(b.col(deltaWidth-1));	// copy first col of image to the last col of matrix b
	for (int j=0; j< deltaWidth-2;j++){
	  b.at<double>(0,1+j) = (b.at<double>(0,deltaWidth-1)*j+b.at<double>(0,0)*(deltaWidth-3-j))/(deltaWidth-3);
	  b.at<double>(imgLastRow-1,1+j) = (b.at<double>(imgLastRow-1,deltaWidth-1)*j
			  +b.at<double>(imgLastRow-1,0)*(deltaWidth-3-j))/(deltaWidth-3);
	}
	this->solveMinLaPlacion(b);

	// calculate the lower right square c
	Mat c  = Mat::zeros(deltaHeight, deltaWidth, CV_64FC1);
	b.row(imgLastRow-1).copyTo(c.row(0)); // copy last row of b to first row of matrix c
	b.row(0).copyTo(c.row(deltaHeight-1));	// copy first row of b to the last row of matrix c
	a.col(imgLastCol-1).copyTo(c.col(0)); // copy last col of a to the first col of matrix c
	a.col(0).copyTo(c.col(deltaWidth-1));	// copy first col of a to the last col of matrix c
	this->solveMinLaPlacion(c);


	// combine plane with a +b +c
	hconcat(planes[i],b(Rect(1,0,deltaWidth-2,imgLastRow)),planes[i]);
	hconcat(a(Rect(0,1,imgLastCol, deltaHeight-2)), c(Rect(1,1,deltaWidth-2, deltaHeight-2)),a);
	planes[i].push_back(a);
  }

  merge(planes, channels, image);

  //Set attribute padded image
  this->imagePadded = this->image;
}

/**
 * Sets the input mat to the original one channel image attribute . If required the mat is converted to greyscale.
 * If the boolean useGammaCorrection is set, then gamma channel of the matrix is first corrected with the value stored
 * in the attribute gammaCorrection.
 *
 * @param mat	The matrix to which the imageOriginalOneChannel is set.
 * @param useGammaCorrection  Flag to indicate whether gamma corrections needs to be applied to the mat before setting it as attribute
 * @throws domain_error  if the depth of the mat is not 64 bits.
 *
 */
void BlurredImage::setImageOriginalOneChannel(cv::Mat mat, bool useGammaCorrection) throw (domain_error){

  // check whether depth of image orignal is 64 float
  if (mat.depth() != 6){
    throw domain_error("BlurredImage; setImageOriginalOneChannel: depth om mat is not 64 float ");
  }


  if (mat.channels()>=3){
      mat.convertTo(imageOriginalOneChannel, CV_32F);
      cv::cvtColor(imageOriginalOneChannel, imageOriginalOneChannel, CV_BGR2GRAY); // convert to gray scale and copy to imageOriginal
      imageOriginalOneChannel.convertTo(imageOriginalOneChannel, CV_64F);
  }
  else {
      if (mat.channels()==2) {
          Mat temp[2];
          split(mat,temp);
          temp[0].copyTo(imageOriginalOneChannel);
      }
      else{
          mat.copyTo(imageOriginalOneChannel);
      }
  }

  // adjust for gamma correction
  if (gammaCorrection != 1 && useGammaCorrection) {
      pow(imageOriginalOneChannel, gammaCorrection, imageOriginalOneChannel);
  }
}

/**
 *  Determines the next largest optimal size for FFT calculations given the minimum required size.
 *  @param size  minimum required size
 *  @return next largest size with is an optimal size for fft.
 */
gint BlurredImage::getOptimalSize(gint size){
  return this->optimalSize[size];
}



/**
 * Sets the working copy of the blurred image to the image without padding.
 */
void BlurredImage::setImageNoPadding(){
	this->image = this->imagePadded(Rect(0,0, this->sizeRescaled.width, this->sizeRescaled.height));
}

/**
 * Sets the working copy of the blurred image to the image with padding.
 */
void BlurredImage::setImagePadding(){
	this->image = this->imagePadded;
}

/**
 * return the orignal blurred image converted to one channel and possible inclusing gamma correction.
 * Note that the conversion can either be a conversion to greyscale
 * or it can be just a snapshot of onechannel of the originalImage.
 *
 * @return  The orignal blurred image converted to one channel.
 */
cv::Size BlurredImage::getOriginalOneChannelSize(){
	return this->imageOriginalOneChannel.size();
}


/**
 * Returns the applied gammaCorrection
 *
 * @return applied gammaCorrection
 */
gdouble BlurredImage::getGammaCorrection(){
	return this->gammaCorrection;
}

/**
 * Sets the gammaCorrection factor
 *
 * @param gammaCorrection  The factor to be used for the gamma correction.
 */
void BlurredImage::setGammaCorrection(gdouble gammaCorrection) {
	this->gammaCorrection = gammaCorrection;
}

/**
 * Returns the orignal blurred image.
 *
 * @return  The original blurred image.
 */
cv::Mat BlurredImage::getImageOriginal(){
	return this->imageOriginal;
}

/**
 * Returns the size of the rescaled blurred image.
 *
 * @return Size to which the blurred image is rescaled.
 */
Size BlurredImage::getSizeRescaled(){
	return this->sizeRescaled;
}

/**
 * Sets the size to which the blurred image needs to be rescaled.
 *
 * @param size  The size to which the blurred image needs to be rescaled.
 */
void BlurredImage::setSizeRescaled(Size size){
	this->sizeRescaled = size;
}

/**
 * Sets the Mat mat to the imagePadded attribute.
 * @param mat The matrix to which the imagePadded attribute is set.
 */
void BlurredImage::setImagePadded(Mat mat){
	this->imagePadded = mat;
}

/**
 * Rescales the image down with the factor given using a scaling kernel
 *
 * @param image The image that will be resized.
 * @param factor The reduction factor. Note that factor should be between 1 and 0
 * @return the rescaled image.
 * @throws domain_error  if factor is not between 1 and 0.
 *
 */
Mat BlurredImage::downSmpImg(cv::Mat image, double factor) throw (domain_error){

  // validate input
  if(factor <= 0 || factor > 1){
	throw domain_error("BlurredImage; downSmpIng: factor should be between 1 and 0 ");
  }
  // only resize if factor <> 1
  if(factor==1){
	return image;
  }

  // calculate kernel for convolution
  double sig = 1/pi*factor;
  std::vector<double> sf;
  std::vector<double> csf;
  double g0=0;
  double sumsf=0;

  for (int i = 0; i < 101; i++){
	g0 = (i-50)*2*pi;
	double sfv = exp(-0.5*pow(g0,2)*pow(sig,2));
	sf.push_back(sfv);
	sumsf = sumsf + sfv;
  }


  // calculate cummulative sum for thresholding and normalise sf
  double csfCounter = 0;
  for (int i = 0; i < 101; i++){
	sf[i] = sf[i] / sumsf; // normalize
	// create cummulative array
	if (i < 51){  // cummulative array needs to be mirror at i = 50
	  csfCounter = csfCounter + sf[i];
	  csf.push_back(csfCounter);
	} else {
	  csf.push_back(csf[100-i]);
	}
  }

  double temp = 0;
  for (int i = 0; i<101; i++){
	temp = temp + csf[i];
  }


  // remove values in sf with corresponding CSF value below threshold 0.5
  for (int i = 0; i < 101; i++){
    if (csf[100-i] < 0.05){   // start at end to keep index equal to i at the fornt part of sf
    	sf.erase(sf.begin()+100-i);
    }
  }

  // convolve the image s with vector sf in both horizontal and vertical direction
  Mat sfMat = Mat(sf);
  Mat imageConvolved;
  Point anchor(0,0); // kernel anchor point at (0,0)

  filter2D(image, imageConvolved, -1, sf, anchor); // horizontal convolution
  filter2D(imageConvolved, imageConvolved, -1, sfMat, anchor); // vertical convolution
  // remove border with incomplete convolution
  imageConvolved = imageConvolved(Rect(0,0,image.cols - sf.size()+1, image.rows - sf.size()+1));

  // resize convolved image with bilinear interpolation
  Mat res = interp2(imageConvolved, factor);
  return res;
}

/**
 * 2 dimensional interpolation of the Mat src which is downscaled with a factor scale.
 * The scale should be greater then zero.
 *
 * @param src   The input mat which is scaled via 2-d interpolation
 * @param scale	The factor used for scaling down the image
 * @throw domain_error 	if scale is not greater then zero
 *
 */
Mat BlurredImage::interp2(const Mat& src, gdouble scale) throw (domain_error){

  if(scale <= 0 ){
	  throw domain_error("BlurredImage; interp2: scale factor needs to be greater then zero.");
  }

  Mat rescaled;
  Mat temp;
  src.copyTo(temp);
  if(scale==1){
      src.copyTo(rescaled);
  }
  else{
      gdouble step = 1/scale;
      gint xsize;
      gint dim, i, j;
      gint coord1, coord2;
      gdouble weight1, weight2;
      gdouble xloc, distance;
      gdouble newval;
      for (dim = 1; dim < 3; dim++){
          xsize = floor(scale * (temp.cols-1)) + 1;
          rescaled = Mat::zeros(temp.rows, xsize, temp.type());
          for(i = 0; i < xsize; i++){
              xloc = i * step;
              distance = xloc - floor(xloc);
              coord1 = floor(xloc);
              coord2 = coord1 + 1;
              weight1 = 1 - distance;
              weight2 = 1 - weight1;
              for (j = 0; j < temp.rows; j++){
                  newval = weight1 * temp.at<double>(j, coord1);
                  newval = newval + weight2 * temp.at<double>(j, coord2);
                  rescaled.at<gdouble>(j,i) = newval;
              }
          }
          transpose(rescaled, temp);
      }
      rescaled = temp;
  }
  return rescaled;
}


/**
 * Method used by the wrapBoundaryLiu method for determining the values for the additional rows and colums.
 * Given a matrix which has only values for the edge coordinates, it calculates a value for all the internal coordinates based
 * on filtering in the spacial domain via discrete sinus transformations
 *
 * @param image  The input image which has only values for the coordinates along the edges
 */
void BlurredImage::solveMinLaPlacion(Mat& image){
  Size size = image.size();

  // only the border value can be non zero
  image(Rect(1,1,size.width-2,size.height-2)).setTo(Scalar(0));

  //set second outer border
  Mat f_bp = Mat::zeros(size, CV_64FC1);
  for (int i = 1; i < size.width-1; i++) {
	  f_bp.at<double>(1,i,0) = image.at<double>(0,i,0);
	  f_bp.at<double>(size.height-2,i,0) = image.at<double>(size.height-1,i,0);
  }
  for (int i = 1; i < size.height-1; i++) {
	  f_bp.at<double>(i,1,0) = f_bp.at<double>(i,1,0) + image.at<double>(i,0,0);
	  f_bp.at<double>(i, size.width-2,0) = f_bp.at<double>(i, size.width-2,0)+ image.at<double>(i, size.width-1,0);
  }

  // compute discrete sine transformation
  Mat f2 = -1 * f_bp(Rect(1,1,size.width-2,size.height-2));
  Mat tt = this->dst(f2);
  Mat f2sin = this->dst((tt.t()));
  f2sin = f2sin.t();


  //create meshgrid
  vector<double> xRow, yCol;
  for (double i = 1; i < size.width-1; i = i+1){
	  xRow.push_back(i);
  }
  for (double i = 1; i< size.height-1; i=i+1){
	  yCol.push_back(i);
  }
  Mat gridX, gridY;
  repeat(Mat(xRow).t(), Mat(yCol).total(), 1, gridX);
  repeat(Mat(yCol), 1, Mat(xRow).total(), gridY);


  // calculate denominator and divice
  Mat denom = cosPerElement(pi*gridX/(size.width - 1))*2-2;
  denom = denom + (cosPerElement(pi*gridY/(size.height-1))*2-2);
  Mat f3;
  divide(f2sin, denom,f3);

  // compute Inverse Sine Transform
  tt= idst(f3);
  Mat img_tt = idst(tt.t()).t();


  // paste result into middle of input image
  img_tt.copyTo(image(Rect(1,1,img_tt.cols, img_tt.rows)));

};

/**
 * Convert the mat to depth 64 float and sets it to the attribute imageOriginal
 *
 * @param mat Image that is converted and set to the attributed
 *
 * @throw domain_error if matrix values of the imageoriginal are not between 0 en 1.
 */
void BlurredImage::setImageOriginal(cv::Mat mat) throw (domain_error){
  // check if mat has depth 64 float, otherwise convert to 64 float
  if (mat.depth() == CV_32F){
	  mat.convertTo(imageOriginal, CV_64FC1);
  }
  else if(mat.depth() != CV_64F){
	  // all other depth are assumed to have pixel values between 1 and 0
	  mat.convertTo(imageOriginal, CV_64FC1, 1.0/255);
  } else {
	  // if mat is already 64 float then assign to attribute imageOriginal
	  imageOriginal = mat;
  }

  // check if values are between 0 and 1
  double min, max;
  minMaxLoc(imageOriginal, &min, &max);
  if(min < 0 || max > 1){
	throw domain_error("BlurredImage.setImageOriginal: values of the input mat exceed the boundaries.");
  }
}


/**
 * returns the original image. Note that this image is converted to 64 bits color depth
 *
 * @return  Original image in 64 bits color.
 */
Mat BlurredImage::getImageOriginalOneChannel(){
	return this->imageOriginalOneChannel;
}

/** returns the paddedImage attribute.
 *
 * @return padded image
 *
 */
Mat BlurredImage::getImagePadded(){
	return this->imagePadded;
}

