/*
 * FrameworkController.cpp
 *
 * @author  E. van Schaijk
 * @author  T. Poppe
 * @version  1.0
 * @since  July, 2016
 *
 * This code implements the image deblurring algorithm described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 *
 */

#include "FrameworkController.hpp"

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/opencv.hpp>
#include <iostream>
#include <stdexcept>
#include <opencv2/core/core_c.h>

#include "Kernel.hpp"
#include "LatentImage.hpp"
#include "BlurredImage.hpp"
#include "../textDeblurPanGui/DeblurPanPlugin.h"
extern "C" {
  #include <log4c.h>

}

using namespace cv;
using namespace std;

log4c_category_t* mycat = log4c_category_get("deblur.pan.frameworkcontroller");

/**
 * Create a new instance of the FrameworkController class.
 */
FrameworkController::FrameworkController(void){
  this->initializeMaxKernelSize();
};

/**
 * Executes a uniform motion deblur operation on the blurred image
 * specified by the inBlurred parameter.
 * The deblur algorithm is an implementation of the algorithm of Pan et al.
 * @param param  the values for the deblur parameters used by the algorithm.
 * @param inBlurred  a reference to a Mat object representing the blurred image.
 * @param outKernel  a reference to a Mat object in which the calculated kernel will be stored.
 * @param outIntermediate  a reference to a Mat object in which the intermediate latent image will be stored.
 * @param outLatent  a reference to a Mat object in which the deblurred image will be stored.
 * @throws domain_error  if param.kernelSize < 3
 * @throws domain_error  if param.kernelSize > getMaxKernelSize()
 * @throws domain_error  if param.gammaCorrection < 1
 * @throws domain_error  if param.lambdaPixel < 0
 * @throws domain_error  if param.lambdaGrad = 0 or lambdaGrad < 0
 * @throws domain_error  if param.numberIterations = 0
 * @throws domain_error  if param.lambdaTV < 0
 * @throws domain_error  if param.lambdaL0 < 0
 * @throws domain_error  if param.weightRing < 0
 * @throws domain_error  if inBlurred width < 5 or inBlurred height < 5
 * @throws domain_error  if inBlurred width > 4096 and inBlurred height > 4096
 */
void FrameworkController::deblur(DeblurParameters& param, Mat& inBlurred, Mat& outKernel, Mat& outIntermediate, Mat& outLatent) throw (domain_error)
{
  string methodId = "FrameworkController.deblur";
  string message;
  if(param.kernelSize < 3){
      message = "kernel size must be at least 3";
      throw domain_error(methodId + ": " + message);
  }
  if(inBlurred.cols > 4096 && inBlurred.rows > 4096){
      message = "blurred width and blurred height must not both exceed 4096";
      throw domain_error(methodId + ": " + message);
  }
  Size blurredSize = Size(inBlurred.cols, inBlurred.rows);
  if(param.kernelSize > this->getMaxKernelSize(blurredSize)){
      message = "the kernel is too large compared to the size of the blurred image";
      throw domain_error(methodId + ": " + message);
  }
  if(param.numberIterations == 0){
      message = "number of iterations must be at least 1";
      throw domain_error(methodId + ": " + message);
  }
  if(param.gammaCorrection < 1){
      message = "gammaCorrection must be at least 1";
      throw domain_error(methodId + ": " + message);
  }
  if(param.lambdaTV < 0 || param.lambdaL0 < 0 || param.weightRing < 0){
      message = "lambdaTV, lambdaL0 and weightRing must not be negative";
      throw domain_error(methodId + ": " + message);
  }

  deblurParameters = param;
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "Start deblurring");

  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE,
          "Input kernel size %dx%d image size %dx%d",
		  deblurParameters.kernelSize, deblurParameters.kernelSize, inBlurred.cols, inBlurred.rows);

  guint kernelSizeDef = (deblurParameters.kernelSize % 2 == 0)? deblurParameters.kernelSize - 1: deblurParameters.kernelSize;

  kernel = Kernel(kernelSizeDef);
  blurred = BlurredImage(inBlurred, deblurParameters.gammaCorrection);
  latent = LatentImage();

  gdouble ret = sqrt(0.5);
  gdouble retv;
  gint maxItr = floor(log(5/(double)kernelSizeDef)/log(ret));
  gint numSteps = (maxItr<0) ? 1 : maxItr + 1;
  gint step, iteration, ksize;
  gdouble lambdaPixel = deblurParameters.lambdaPixel;
  gdouble lambdaGrad = deblurParameters.lambdaGrad;
  gdouble progress = 0.0;

  for (step = numSteps ; step > 0 ; step--)
  {
    retv = pow(ret, step-1);
    ksize = ceil(kernelSizeDef * retv);
    ksize = (ksize % 2 == 0)? ksize + 1: ksize;

    blurred.resizeForFFT( retv, ksize );

    if(step==numSteps){
        kernel.initialize(ksize);
        blurred.setImageNoPadding();
        gdouble th = kernel.initializeThreshold(blurred);
        if(th != 0 && th < lambdaGrad/10){
            lambdaGrad = th;
            lambdaPixel = th;
        }
    }
    else{
        kernel.resizeForStep(Size(ksize, ksize));
    }

    gint bsize[] = {blurred.getSizeRescaled().width, blurred.getSizeRescaled().height};
    log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE,
        "Processing scale %d/%d kernel size %dx%d image size %dx%d",
        step, numSteps, ksize, ksize, bsize[0], bsize[1]);
    cout << "Processing scale " << step << "/" << numSteps <<
        " kernel size " << ksize << "x" << ksize <<
        " image size " << bsize[0] << "x" << bsize[1] << endl;

    latent.initialize(bsize);

    for ( iteration = 1; iteration < deblurParameters.numberIterations + 1; iteration++ )
    {
        log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "Processing iteration %d/%d",
            iteration, deblurParameters.numberIterations);

        latent.calculateLatent(kernel, blurred, lambdaPixel, lambdaGrad);

        kernel.calculateKernel(blurred, latent, Size(ksize, ksize));

        // update parameters
        if (lambdaPixel != 0 ){
            lambdaPixel = max(lambdaPixel/1.1, 1e-4);
        }
        if (lambdaGrad != 0 ){
            lambdaGrad = max(lambdaGrad/1.1, 1e-4);
        }

        progress = 0.01 * estimateProgress(numSteps, deblurParameters.numberIterations, numSteps+1-step, iteration, deblurParameters.weightRing);
        this->notify(progress);
    }
    kernel.adjustPSFCenter();
  }
  kernel.normalize(20);
  Mat intermediate;
  intermediate = latent.getImage();
  intermediate.setTo(0, intermediate < 0);
  intermediate.setTo(1, intermediate > 1);

  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "Start Final Deconvolution");

  latent.calculateLaPlacian(kernel, blurred, deblurParameters.lambdaTV, deblurParameters.lambdaL0, deblurParameters.weightRing);

  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "End Final Deconvolution");

  outKernel = convertImage(kernel.getImage(), inBlurred.depth(), true);
  outIntermediate = convertImage(intermediate, inBlurred.depth(), false);
  outLatent = convertImage(latent.getImage(), inBlurred.depth(), false);

  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE,
      "Results: kernel size %dx%d intermediate size %dx%d latent size %dx%d",
      outKernel.cols, outKernel.rows, outIntermediate.cols, outIntermediate.rows,
      outLatent.cols, outLatent.rows);
}

/**
 * Estimates the progress of the deblur operation after
 * finishing the given iteration within the given step.
 * The progress estimation is based on ret = sqrt(0.5).
 * Step counts from 1 to steps.
 * Iteration counts from 1 to iterations.
 * @param steps  The number of steps of the deblur operation
 * @param iterations The number of iterations within each step
 * @param step  The current step
 * @param iteration  The current iteration
 * @param ringingRemoval  Indicates whether ringing removal is applied
 * @return  The progress of the deblur operation
 * @throws domain_error  if steps < 1
 * @throws domain_error  if iterations < 1
 * @throws domain_error  if step < 1 or step > steps
 * @throws domain_error  if iteration < 1 or iteration > iterations
 */
gdouble FrameworkController::estimateProgress(gint steps, gint iterations, gint step, gint iteration, gboolean ringingRemoval) throw (domain_error)
{
  string methodId = "FrameworkController.estimateProgress";
  string message;
  if(steps < 1){
      message = "steps must be at least 1";
      throw domain_error(methodId + ": " + message);
  }
  if(iterations < 1){
      message = "iterations must be at least 1";
      throw domain_error(methodId + ": " + message);
  }
  if(step < 1 || step > steps){
      message = "step must be in the range from 1 to steps";
      throw domain_error(methodId + ": " + message);
  }
  if(iteration < 1 || iteration > iterations){
      message = "iteration must be in the range from 1 to iterations";
      throw domain_error(methodId + ": " + message);
  }

  gdouble ctf = 2.4;
  gdouble ftc = 1.0/ctf;

  //observed reference values for steps = 1 and iterations = 1
  gdouble fdRef = ringingRemoval ? 30.0 : 10.0;
  gdouble keRef = 100.0 - fdRef;

  //calculate percentage of processing time needed for kernel estimation
  gdouble keLength = (1.0 - pow(ftc,steps))/(1.0 - ftc);
  gdouble keFactor = iterations * keLength;
  gdouble kePercent = 100.0 * keFactor * keRef / (keFactor * keRef + fdRef);

  //calculate progress
  gdouble stepStart = (pow(ftc, (steps- step + 1)) - pow(ftc,steps))/(1.0 - ftc);
  gdouble fraction = (gdouble) iteration / iterations;
  gdouble stepAdd = fraction * (pow(ftc,steps-step));
  gdouble progress = kePercent * (stepStart + stepAdd)/keLength;

  return progress;
}

/**
 * Initializes the attribute MaxKernelSize by creating a lookup array.
 * Assumes the kernel size is odd and the kernel size is at least 3.
 */
void FrameworkController::initializeMaxKernelSize(void){

  gint arraysize = sizeof(this->maxKernelSize)/sizeof(this->maxKernelSize[0]);
  gdouble ret = sqrt(0.5);
  gdouble retv;
  gint maxItr, kmin, bmin;
  gint bminErr = 3;
  gint i, s;

  gint ksize = 0;
  this->maxKernelSize[0] = 0;
  for (i = 1; i < arraysize; i++){
      ksize = ksize + 1;
      maxItr = max((gint)floor(log(5/(double)ksize)/log(ret)),0);
      retv = pow(ret, maxItr);
      if(maxItr == 0){
          kmin = ksize;
          bmin = i;
      }
      else{
          kmin = ceil(ksize * retv);
          kmin = (kmin % 2 == 0)? kmin + 1: kmin;
          bmin = ceil(i * retv) - bminErr;
      }
      if(kmin >= bmin){
          ksize = ksize - 1;
      }
      if(ksize < 3){
          this->maxKernelSize[i] = 0;
      }
      else{
          s = (ksize % 2 == 0)? ksize - 1: ksize;
          this->maxKernelSize[i] = s;
      }
  }
}

/**
 * Returns the maximum kernel size for the given size of the blurred image.
 * The maximum kernel size is searched across the lookup array maxkernelSize.
 * @param blurredSize	The size of the blurred image.
 * @return		The maximum size of the kernel.
 * @throws domain_error  if blurredSize width < 0 or blurredSize height < 0
 * @throws domain_error  if blurredSize width > 4096 and blurredSize height > 4096
 */
gint FrameworkController::getMaxKernelSize(Size blurredSize) throw (domain_error)
{
  string methodId = "FrameworkController.getMaxKernelSize";
  string message;

  if (blurredSize.width < 0  || blurredSize.height < 0){
      message = "blurredSize must not have any negative values";
      throw domain_error(methodId + ": " + message);
  }

  gint maxbsize = sizeof(this->maxKernelSize)/sizeof(this->maxKernelSize[0])-1;
  gint bsize = min(blurredSize.width, blurredSize.height);

  if (bsize > maxbsize){
      ostringstream msg;
      msg << "blurred width and blurred height must not both exceed " << maxbsize;
      throw domain_error(methodId + ": " + msg.str());
  }
  return this->maxKernelSize[bsize];
}

/**
 * Converts the src image depth to the depth specified by the
 * input parameter depth.
 * If boolean stretch is true than the values are stretched
 * to the upper bound corresponding to the image depth.
 * @param src  the image to be converted
 * @param depth  the target depth of the image
 * @param stretch  indicates whether the values of src must be
 *                 stretched to the upper bound
 * @return the src image converted to the target depth
 * @throws domain_error  if src depth < CV_32F
 * @throws domain_error  if src maximum value > 1
 * @throws domain_error  if src minimum value < 0
 * @throws domain_error  if depth > CV_64F
 */
Mat FrameworkController::convertImage(const Mat& src, gint depth, gboolean stretch) throw (domain_error)
{
  string methodId = "FrameworkController.convertImage";
  string message;
  if(src.depth() < CV_32F){
      message = "src depth must be at least CV_32F";
      throw domain_error(methodId + ": " + message);
  }
  if(depth > CV_64F){
      message = "depth must not exceed CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  gdouble minVal, maxVal;
  minMaxIdx(src, &minVal, &maxVal);
  if(minVal < 0 || maxVal > 1){
      message = "src must have values between 0 and 1";
      throw domain_error(methodId + ": " + message);
  }

  Mat converted;
  src.copyTo(converted);
  gdouble scale = stretch ? 1.0/maxVal : 1.0;

  if(depth != src.depth()){
      if(depth < CV_32F){
          converted.convertTo(converted, depth, 255.0 * scale);
      }
      else{
          converted.convertTo(converted, depth, scale);
      }
  }
  return converted;
}





