/*
 * FrameworkController.hpp
 *
 * @author  E. van Schaijk
 * @author  T. Poppe
 * @version  1.0
 * @since  July, 2016
 *
 * This code implements the image deblurring algorithm described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 *
 */

#ifndef TEXTDEBLURPAN_FRAMEWORKCONTROLLER_HPP_
#define TEXTDEBLURPAN_FRAMEWORKCONTROLLER_HPP_

#include <stdexcept>

#include "Kernel.hpp"
#include "LatentImage.hpp"
#include "Observable.hpp"
#include "BlurredImage.hpp"
#include "../textDeblurPanGui/DeblurPanPlugin.h"

extern "C" {
  #include <opencv2/core/core_c.h>
}

/*
 * Struct for storing the deblur parameter settings
 */
struct DeblurParameters
{
  guint kernelSize; 		///< Size of the kernel which is used for the kernel estimation.
  gdouble gammaCorrection; 	///< gamma factor used for applying gamma correction.
  gdouble lambdaPixel;		///< lambda Pixel coefficient, determines the weight for the L0 pixel intensity prior.
  gdouble lambdaGrad;		///< lambda Grad coefficient, determines the weight for the L0 gradient prior.
  guint numberIterations;	///< number of iterations of the inner loop.
  gdouble lambdaTV;			///< lambda TV coefficient used for the Laplacian prior during the non blind deconvolution.
  gdouble lambdaL0;			///< lambda l0 grad coefficient, determines the weight for the L0 gradient prior during the non blind deconvolution.
  gdouble weightRing;		///< Weight factor applied for ringing removal via bilateral filtering.
} ;

/**
 * Class used as central controller which implements the uniform motion deblur algorithm of Pan et al.
 */
class FrameworkController: public Observable {

private:
  FrameworkController(void);  //Disallowed
  FrameworkController(const FrameworkController&); //Disallowed
  FrameworkController& operator=(const FrameworkController&); //Disallowed
  Kernel kernel = Kernel(10); 	///< attribute representing the kernel.
  LatentImage latent;			///< attribute representing the latent image.
  BlurredImage blurred;			///< attribute representing the blurred image.
  gint maxKernelSize[4097]; //inclusief 0!
  void initializeMaxKernelSize(void);
  DeblurParameters deblurParameters;

public:
  void deblur(DeblurParameters& param, cv::Mat& inBlurred, cv::Mat& outKernel, cv::Mat& outIntermediate, cv::Mat& outLatent) throw (std::domain_error);
  gint getMaxKernelSize(cv::Size blurredSize) throw (std::domain_error);

  /**
   * return the frameworkController object. If required a new object is created.
   * @return	frameworkController object.
   */
  static FrameworkController& getInstance(void){
    static FrameworkController instance;
    return instance;
  }

  //nog private maken
  gdouble estimateProgress(gint steps, gint iterations, gint step, gint iteration, gboolean ringingRemoval) throw (std::domain_error);
  cv::Mat convertImage(const cv::Mat& src, gint depth, gboolean stretch) throw (std::domain_error);
};

#endif /* TEXTDEBLURPAN_FRAMEWORKCONTROLLER_HPP_ */
