/**
 * ImageParameter.cpp
 *
 * @author  E. van Schaijk
 * @author  T. Poppe
 * @version  1.0
 * @since  July, 2016
 *
 * This code implements the image deblurring algorithm described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 *
 */

#include "ImageParameter.hpp"

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <iostream>
#include <stdexcept>

using namespace cv;
using namespace std;

/**
 * Creates a new ImageParameter.
 */
ImageParameter::ImageParameter() {
}

/**
 * Returns the width of this ImageParameter's image.
 * If this ImageParameter's image is empty then 0 is returned.
 * @return  the width of this ImageParameter's image
 */
gint ImageParameter::getCurrentWidth(void)
{
  return this->image.size.p[1]; //cols
}

/**
 * Returns the height of this ImageParameter's image.
 * If this ImageParameter's image is empty then 0 is returned.
 * @return  the height of this ImageParameter's image
 */
gint ImageParameter::getCurrentHeight(void)
{
  return this->image.size.p[0]; //rows
}

/**
 * Returns a copy of this ImageParameter's image.
 * If this ImageParameter's image is empty then an empty image
 * is returned.
 * @return  a copy of this ImageParameter's image
 */
Mat ImageParameter::getImage(void)
{
  Mat m;
  if(this->image.data) this->image.copyTo(m);
  return m;
}

/**
 * Sets the image of this ImageParameter.
 * @param image  the image to be set
 */
void ImageParameter::setImage(cv::Mat image)
{
  this->image = image;
}

/**
 * Returns the x-gradients of this ImageParameter's image.
 * The x-gradient per pixel is calculated by extracting the
 * value of the right neighbor pixel from the current pixel's
 * value.
 * The pixels in the top-most row and the right-most column
 * of the source image will be neglected. Therefore, the size
 * of the returned gradient map is one row and one column
 * smaller than the size of the source image.
 * @return  the x-gradients of this ImageParameter's image
 * @throws domain_error  if image rows < 2 or image cols < 2
 */
Mat ImageParameter::getXGrad(void) throw (domain_error)
{
  Mat xgrad;
  gint dx_in[2][2] = {{-1, 1},{0, 0}}; //derivate filter
  Mat dx = Mat(2,2,CV_32S,&dx_in);
  xgrad = conv2(dx); //throws domain_error
  return xgrad;
}

/**
 * Returns the y-gradients of this ImageParameter's image.
 * The y-gradient per pixel is calculated by extracting the
 * value of the bottom neighbor pixel from the current pixel's
 * value.
 * The pixels in the bottom-most row and the left-most column
 * of the source image will be neglected. Therefore, the size
 * of the returned gradient map is one row and one column
 * smaller than the size of the source image.
 * @return  the y-gradients of this ImageParameter's image
 * @throws domain_error  if image rows < 2 or image cols < 2
 */
Mat ImageParameter::getYGrad(void) throw (domain_error)
{
  Mat ygrad;
  gint dy_in[2][2] = {{-1, 0},{1, 0}}; //derivate filter
  Mat dy = Mat(2,2,CV_32S,&dy_in);
  ygrad = conv2(dy); //throws domain_error
  return ygrad;
}

/**
 * Calculate the 2d FFT (also known as DFT) of the matrix src.
 *
 * Input: Matrix with 1 channels containing the real values or a matrix with 2 channels with Real and Img values.
 * Output: the FFT transform of src.
 */
Mat ImageParameter::fft2(const Mat& src){

  // validatie current image
  if (src.empty()){
	  throw domain_error("No image is set");
  }
  if (src.channels() > 2 ){
	  throw domain_error("FFT can only be calculated for images with 1  or 2 channels");
  }

  Mat complexMat;
  if (src.channels() == 1 ){
	  Mat planes[] = {Mat_<double>(src), Mat::zeros(src.size(), CV_64F)};
	  merge(planes, 2, complexMat);
	  planes[0].release();
	  planes[1].release();
  } else {
	  complexMat = src;
  }

  Mat res;
  dft(complexMat, res);

  return res;
}

/**
 * Calculate the inverse 2d fft of input src.
 *
 * input: a matrix src containing 2 channels one with the real values and a second one with the img values.
 * Output: the real values of the inverse fft.
 */
Mat ImageParameter::ifft2(const Mat& src){

  // validatie current image
  if (src.empty()){
	  throw domain_error("No image is set");
  }
  if (src.channels() != 2 ){
	  throw domain_error("inverse FFT can only be calculated for images with 2 channels (RE , IMG) ");
  }

  cv::Mat inverseTransform;
  cv::dft(src, inverseTransform, cv::DFT_INVERSE|cv::DFT_REAL_OUTPUT);

  // normalize to orignal values
  int totalElem = inverseTransform.total();
  Mat res = inverseTransform / totalElem;
  Mat res2;
  res.copyTo(res2);


  return res2;
}

/**
 * Calculate the 2d FFT (also known as DFT) of the matrix src and returns the complex conjugate of the FFT
 *
 * Input: Matrix with 1 channels containing the real values or a matrix with 2 channels with Real and Img values.
 * Output: complex conjugate of the FFT transform of src.
 */
Mat ImageParameter::conjFFT(const Mat& src) {


  Mat fftRes = fft2(src);
  Mat planes[2];
  cv::split(fftRes, planes); // planes[0] = Re(DFT(I), planes[1] = Im(DFT(I))
  fftRes.release();

  planes[1] = -planes[1];
  merge(planes, 2, fftRes);


  planes[0].release();
  planes[1].release();

  Mat fftRes2;
  fftRes.copyTo(fftRes2);

  return fftRes2;
}

/**
 * Creates the optical transfer function (OTF) from the given point
 * spread function (PSF).
 * The OTF is defined as the Fourier Transform of the PSF. To ensure
 * that the OTF is not altered because of PSF off-centering, the PSF
 * is first zero-padded to match the dimensions specified by size, then
 * circularly shifted until the central pixel reaches (0,0) position.
 * @param psf  the input point spread function
 * @param size  the size of the optical transfer function to be created
 * @return  the optical transfer function otf created from the given
 *          point spread function psf
 * @throws domain_error  if psf is empty
 * @throws domain_error  if psf channels > 1
 * @throws domain_error  if size width < psf cols or size height < psf rows
 */
Mat ImageParameter::psf2otf(const Mat& psf, Size size) throw (std::domain_error)
{
  string methodId = "ImageParameter.psf2otf";
  string message;
  if(psf.empty()){
      message = "psf must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  if(psf.channels() > 1){
      message = "psf must have no more than one channel";
      throw domain_error(methodId + ": " + message);
  }
  if(size.width < psf.cols || size.height < psf.rows){
      message = "size must be at least the size of the psf";
      throw domain_error(methodId + ": " + message);
  }

  //apply padding
  Mat psf_padded;
  copyMakeBorder(psf, psf_padded, 0, size.height - psf.rows, 0, size.width - psf.cols, BORDER_CONSTANT, 0 );

  //apply circular shift
  gint shift_rows = (psf.rows % 2 == 0) ? psf.rows/2 : (psf.rows + 1)/2;
  shift_rows += size.height - psf.rows;
  gint shift_cols = (psf.cols % 2 == 0) ? psf.cols/2 : (psf.cols + 1)/2;
  shift_cols += size.width - psf.cols;
  Mat psf_shifted = circularShift(psf_padded, shift_rows, shift_cols);

  //apply fft
  Mat otf = fft2(psf_shifted);
  return otf;
}


/**
 * Creates the point spread function (PSF) from the given optical
 * transfer function (OTF).
 * The PSF is defined as the inverse Fourier Transform of the OTF.
 * To center the PSF at the origin, the PSF is first circularly shifted
 * until the (0,0) pixel reaches the central position, then cropped to
 * match the dimensions specified by size.
 * @param otf  the input optical transfer function
 * @param size  the size of the point spread function to be created
 * @return  the point spread function psf created from the given
 *          optical transfer function otf
 * @throws domain_error  if otf is empty
 * @throws domain_error  if otf channels <> 2
 * @throws domain_error  if otf depth < CV_32F
 * @throws domain_error  if size width > otf cols or size height > otf rows
 */
Mat ImageParameter::otf2psf(const cv::Mat& otf, Size size) throw (std::domain_error)
{
  string methodId = "ImageParameter.otf2psf";
  string message;
  if(otf.empty()){
      message = "otf must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  if(otf.channels() != 2){
      message = "otf must have exactly two channels";
      throw domain_error(methodId + ": " + message);
  }
  if(otf.depth() < CV_32F){
      message = "otf depth must be CV_32F or CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  if(size.width > otf.cols || size.height > otf.rows){
      message = "size must not exceed the size of the otf";
      throw domain_error(methodId + ": " + message);
  }

  //apply inverse fft
  Mat ifft = ImageParameter::ifft2(otf);

  //apply circular shift
  gint shift_rows = (size.height % 2 == 0) ? size.height/2 : (size.height - 1)/2;
  gint shift_cols = (size.width % 2 == 0) ? size.width/2 : (size.width - 1)/2;
  Mat shifted = circularShift(ifft, shift_rows, shift_cols);

  //crop image
  Mat psf = shifted(cv::Rect(0, 0, size.width, size.height));
  return psf;
}


/**
 * Computes the 2D convolution of this ImageParameter's
 * image with the given filter.
 * Returns only those parts of the convolution that are
 * computed without the padded edges.
 * @param filter  the convolution filter
 * @return  the 2D convolution of this ImageParameter's
 *          image with the given filter
 * @throws domain_error  if filter is empty
 * @throws domain_error  if filter channels > 1
 * @throws domain_error  if filter depth > CV_32S
 * @throws domain_error  if image rows < filter rows
 *                       or image cols < filter cols
 */
Mat ImageParameter::conv2(const cv::Mat& filter) throw (domain_error)
{
  string methodId = "ImageParameter.conv2";
  string message;
  if(filter.empty()){
      message = "filter must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  if(filter.channels() > 1){
      message = "filter must have no more than one channel";
      throw domain_error(methodId + ": " + message);
  }
  if(filter.depth() > CV_32S) {
      message = "filter depth must be at most CV_32S";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.cols < filter.cols || this->image.rows < filter.rows){
      message = "image is too small for the given filter";
      throw domain_error(methodId + ": " + message);
  }

  Mat filter_copy, image_copy;
  if (filter.depth() < CV_32S){
      filter.convertTo(filter_copy, CV_32S);
  }
  else {
      filter.copyTo(filter_copy);
  }
  if( this->image.depth() < CV_32F){
      this->image.convertTo(image_copy, CV_32F);
  }
  else{
      this->image.copyTo(image_copy);
  }
  Mat fflip;
  Mat result;
  flip(filter_copy, fflip, -1);
  filter2D(image_copy, result, image_copy.depth(), fflip);
  gint c1, c2, r1, r2;
  c1 = fflip.cols %2 == 0 ? fflip.cols/2 : (fflip.cols-1)/2;
  c2 = image_copy.cols - (fflip.cols %2 == 0 ? (fflip.cols/2)-1 : (fflip.cols-1)/2);
  r1 = fflip.rows %2 == 0 ? fflip.rows/2 : (fflip.rows-1)/2;
  r2 = image_copy.rows - (fflip.rows %2 == 0 ? (fflip.rows/2)-1 : (fflip.rows-1)/2);
  result = result.colRange(c1, c2).rowRange(r1, r2);
  return result;
}


/**
 * Circularly shifts the elements in the given image by rows
 * positions in y-direction and cols positions in x-direction.
 * If the shift amount is larger than the length of the corresponding
 * dimension in the image, then the shift circularly wraps to the
 * beginning of that dimension.
 * If the image is a multi-channel image then every separate channel
 * will be shifted.
 * @param  img  the image to be shifted
 * @param  rows  shift amount in y-direction
 * @param  cols  shift amount in x-direction
 * @return  a copy of the image circularly shifted by rows positions
 *          in y-direction and cols positions in x-direction
 * @throws domain_error  if img is empty
 */
Mat ImageParameter::circularShift(const Mat& img, gint rows, gint cols) throw (domain_error)
{
  string methodId = "ImageParameter.circularShift";
  string message;
  if(img.empty()){
      message = "image must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  rows = (rows < 0) ? (img.rows - abs(rows % img.rows)) % img.rows : rows % img.rows;
  cols = (cols < 0) ? (img.cols - abs(cols % img.cols)) % img.cols : cols % img.cols;
  Mat shifted;
  if(rows==0 && cols==0){
      img.copyTo(shifted);
  }
  else{
      shifted = Mat::zeros(img.rows, img.cols, img.type());
      img(Rect(0, 0, img.cols-cols, img.rows-rows)).copyTo(shifted(Rect(cols, rows, img.cols-cols, img.rows-rows)));
      if(cols !=0){
        img(Rect(img.cols-cols, 0, cols, img.rows-rows)).copyTo(shifted(Rect(0, rows, cols, img.rows-rows)));
      }
      if (rows != 0){
    	img(Rect(0, img.rows-rows, img.cols-cols, rows)).copyTo(shifted(Rect(cols, 0, img.cols-cols, rows)));
      }
      if (rows !=0 && cols != 0){
        img(Rect(img.cols-cols, img.rows-rows, cols, rows)).copyTo(shifted(Rect(0, 0, cols, rows)));
      }
  }
  return shifted;
}


/**
 * Calculates the conjugate values of the complex values in the src matrix.
 * @param  src  Matrix containing complex values. The Re and IMG part of the values need to be stored as two separate channels in the matrix
 * @return  	Matrix containing the conjugate values of src. The Re and Imaginary part as stored in two separate channels.
 * @throws domain_error  if src is empty or when the number of channels of src is unequal to 2
 */
Mat ImageParameter::conj(const cv::Mat& src){

  // check source
  if (src.empty()){
	  throw domain_error("No image is set");
  }
  if (src.channels() != 2 ){
	  throw domain_error("conj FFT can only be calculated for images with 2 channels (RE , IMG) ");
  }

  Mat planes[2];
  Mat res;
  cv::split(src, planes); // planes[0] = Re(DFT(I), planes[1] = Im(DFT(I))
  planes[1] = -planes[1];
  merge(planes, 2, res);

  planes[0].release();
  planes[1].release();

  Mat res2;
  res.copyTo(res2);

  return res2;
}

/**
 * Calculates the absolute values of the complex values in the src matrix.
 * The absolute values are calculates as sqrt(Re² + Img²).
 * @param  src  Matrix containing complex values. The Re and IMG part of the values need to be stored as two separate channels in the matrix
 * @return  	Matrix containing the absolute values of src
 * @throws domain_error  if src is empty or when the number of channels of src is unequal to 2
 */
Mat ImageParameter::absModules(const cv::Mat& src){
  // check source
  if (src.empty()){
	  throw domain_error("No image is set");
  }
  if (src.channels() != 2 ){
	  throw domain_error("absModules can only be calculated for images with 2 channels (RE , IMG) ");
  }

  Mat planes[2];
  Mat planes2[2];
  Mat res;
  cv::split(src, planes); // planes[0] = Re(DFT(I), planes[1] = Im(DFT(I))
  pow(planes[0], 2, planes2[0]);
  pow(planes[1], 2, planes2[1]);
  pow(planes2[0]+planes2[1], 0.5, res);

  return res;

}

/**
 * executes a per element multiplication of 2 matrices containing complex numbers.
 * The per element multiplication of matrix A en B is execute as
 * 		Re = Re(A)*Re(B) - Img(A)*Img(B)
 * 		Img = Re(A)*Img(B) + Re(B)*Img(A)
 * @param  srcA  Matrix containing complex values. The Re and IMG part of the values need to be stored as two separate channels in the matrix
 * @param  srcB  Matrix containing complex values. The Re and IMG part of the values need to be stored as two separate channels in the matrix
 * @return  	Matrix containing the multiplicated values of srcA and srcB
 * @throws domain_error  if srcA or srcB is empty or when the number of channels of srcA or srcB is unequal to 2
 */
Mat ImageParameter::mulComplex(const cv::Mat& srcA, const cv::Mat& srcB){
  // check source
  if (srcA.empty() || srcB.empty()){
	  throw domain_error("image A or B is not set");
  }
  if (srcA.channels() != 2 || srcB.channels() != 2){
	  throw domain_error("mulComplex can only be calculated for images with 2 channels (RE , IMG) ");
  }

  Mat planesA[2];
  Mat planesB[2];
  Mat res[2];
  Mat complexRes;
  cv::split(srcA, planesA); // planes[0] = Re(DFT(I), planes[1] = Im(DFT(I))
  cv::split(srcB, planesB); // planes[0] = Re(DFT(I), planes[1] = Im(DFT(I))

  res[0] = planesA[0].mul(planesB[0]) - planesA[1].mul(planesB[1]);
  res[1] = planesA[0].mul(planesB[1]) + planesA[1].mul(planesB[0]);
  merge(res, 2, complexRes);

  return complexRes;

}

/**
 * executes a per element division of a matrix containing complex numbers by a matrix containing floating numbers.
 * The per element division of matrix A en B is execute as
 * 		Re = Re(A)/B
 * 		Img =Img(A) / B
 * @param  srcA  Matrix containing complex values which is used as numerator.
 * 					The Re and IMG part of the values need to be stored as two separate channels in the matrix.
 * @param  srcB  Matrix containing floating values which is used as denominator.
 * @return  	Matrix containing the division values of srcA / srcB
 * @throws domain_error  if srcA or srcB is empty,
 * 						 when the number of channels of srcA is unequal to 2
 * 						 when the number of channels of srcB is unequal to 1
 * 						 when srcB has zero values.
 */
Mat ImageParameter::divComplex(const cv::Mat& srcA, const cv::Mat& srcB){
  // check source
  if (srcA.empty() || srcB.empty()){
	  throw domain_error("image A or B is not set");
  }
  if (srcA.channels() != 2 || srcB.channels() != 1){
	  throw domain_error("mulComplex can only be calculated for images a with 2 channels (RE , IMG) and image b = 1 channels (re) ");
  }

  Mat planesA[2];
  Mat res[2];
  Mat complexRes;
  cv::split(srcA, planesA); // planes[0] = Re(DFT(I), planes[1] = Im(DFT(I))

  // check that denominator is not zero
  int TotalNumberOfPixels = srcB.rows * srcB.cols;
  int ZeroPixels = TotalNumberOfPixels - cv::countNonZero(srcB);
  if (ZeroPixels > 0){
	  throw domain_error("image B has zero pixels");
  }

  res[0] = planesA[0].mul(1/srcB);

  res[1] = planesA[1].mul(1/srcB);

  merge(res, 2, complexRes);

  return complexRes;
}

/**
 * returns are matrix indicating the sign of the individual elements of the matrix src.
 * if the value of the element is <0 then a -1 is returned.
 * if the value of the element is > 0 then a 1 is returned.
 * if the value of the element is =0 then a 0 is returned.
 *
 * @param src	A mat object for which the sign is determined per element.
 * @return		A Mat ojebt with the same size as src indicating for each element the sign.
 *
 */
Mat ImageParameter::signum(Mat& src){
    Mat z = Mat::zeros(src.size(), src.type());
    Mat a = (z < src) / 255;
    Mat b = (src < z) / 255;

    Mat dst;
    addWeighted(a,1.0,b,-1.0,0.0,dst, CV_64F);
    return dst;
}

/**
 * Calculates the discrete sinus transform (also known as DST) of the matrix src.
 *
 * @param src	Matrix with 1 channels containing the real values.
 * @return		The dst transform of src.
 */
Mat ImageParameter::dst(const Mat src){
	Size size = src.size();
	Mat Z = Mat::zeros(Size(size.width,1), CV_64FC1);
	Mat y, imageFlipped;
	Z.copyTo(y);
	y.push_back(src);
	y.push_back(Z);
	flip(src,imageFlipped,0);
	imageFlipped = imageFlipped * -1;
	y.push_back(imageFlipped);

	dft(y.t(), y, DFT_ROWS | DFT_COMPLEX_OUTPUT );
	Mat planesY[2];
	split(y,planesY);
	y = planesY[1].t();

	y = y(Rect(0,1,y.cols ,size.height))/-2;

	return y;
}

/**
 * Calculates the inverse discrete sinus transform (also known as IDST) of the matrix src.
 *
 * @param src	Matrix containing the DST values.
 * @return		The investe DST transform of src also known as the original.
 */
Mat ImageParameter::idst(const Mat src){

  int n = src.rows;
  Mat result = dst(src)*2/(n+1);
  return result;

}

/**
 * executes a per element cosines of a matrix.
 *
 * @param  src  Matrix containing double values of which the cosines is calculated.
 * @return  	Matrix containing the cosines values of src.
 */
Mat ImageParameter::cosPerElement(Mat src){

	Mat result = src.clone();
	for ( int i = 0; i < result.cols; i++){
		for (int j =0; j<result.rows; j++){
			result.at<double>(j,i) = cos(result.at<double>(j,i));
		}
	}
	return result;
}

/**
 * Returns the magnitude map of this ImageParameter's image.
 * The magnitude of a pixel is defined as the sum of squares
 * of its x-gradient and y-gradient. The pixel's x-gradient
 * is calculated relative to its right neighbor pixel. The
 * pixel's y-gradient is calculated relative to its bottom
 * neighbor pixel.
 * @return  the magnitude map of this ImageParameter's image
 * @throws domain_error  if image rows < 2 or image cols < 2
 */
Mat ImageParameter::getMagnitude(void){
  Mat px = this->getXGrad();
  Mat py = this->getYGrad();
  Mat px2, py2;
  pow(px, 2.0, px2);
  pow(py, 2.0, py2);
  Mat pm = px2 + py2;
  return pm;
}
