/*
 * Class Kernel represents the blur kernel.
 *
 * @author  E. van Schaijk
 * @author  T. Poppe
 * @version  1.0
 * @since  July, 2016
 *
 * This code implements the image deblurring algorithm described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 */

#include "Kernel.hpp"

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <iostream>
#include <stdexcept>

using namespace cv;
using namespace std;


/**
 * Creates a new Kernel and initializes the final size
 * of the Kernel's image with the given size.
 * @param size  the final size of this Kernel's image
 * @throws domain_error  if size < 3
 */
Kernel::Kernel(guint size) throw (domain_error)
{
  string methodId = "Kernel.Kernel";
  string message;
  if(size < 2){
      message = "size must be at least 3";
      throw domain_error(methodId + ": " + message);
  }
  this->size = size;
  this->threshold = 0.0;
}

/**
 * Returns the final size of this Kernel's image.
 * @return  the final size of this Kernel's image
 */
guint Kernel::getSize(void) {
  return this->size;
}

/**
 * Initializes this Kernel's image with the given size.
 * Two pixels in the center of the kernel are set to 0.5.
 * All other pixels are set to 0.
 * @param size  the actual size of this Kernel's image
 * @throws domain_error  if size < 3
 * @throws domain_error  if size > this Kernel's size
 */
void Kernel::initialize(guint size) throw (domain_error)
{
  string methodId = "Kernel.initialize";
  string message;
  if(size < 2){
      message = "size must be at least 2";
      throw domain_error(methodId + ": " + message);
  }
  if(size > this->size){
      message = "size must be at most this Kernel's size";
      throw domain_error(methodId + ": " + message);
  }
  gint coord = size%2 == 0 ? (size/2)-1 : ((size-1)/2)-1;
  Mat new_kernel = Mat::zeros(size, size, CV_64FC1);
  new_kernel.at<double>(coord, coord) = 0.5;
  new_kernel.at<double>(coord, coord+1) = 0.5;
  this->image = new_kernel;
}

/**
 * Initializes this Kernel's threshold for discarding the gradients
 * of the latent image during kernel calculation.
 * The initial threshold is calculated from the squared
 * magnitudes of the blurred image.
 * @param blurred  the blurred image
 * @return  the threshold
 * @throws domain_error  if blurred's image channels > 1
 * @throws domain_error  if blurred's image rows < 2 or blurred's image cols < 2
 */
gdouble Kernel::initializeThreshold(ImageParameter& blurred) throw (domain_error)
{
  string methodId = "Kernel.initializeThreshold";
  string message;
  if(blurred.getImage().channels() > 1){
      message = "Blurred's image must have no more than one channel";
      throw domain_error(methodId + ": " + message);
  }

  gint ksize = max(this->getCurrentHeight(), this->getCurrentWidth());
  gfloat th = max(ksize * 20.0, 10.0);
  gfloat mythreshold = 0.0;
  Mat px, py, pm, pd;
  px = blurred.getXGrad();
  py = blurred.getYGrad();
  pm = blurred.getMagnitude();
  pm.convertTo(pm, CV_32F);

  //create mask for px==0 && py==0
  Mat invalid = abs(px) + abs(py);
  invalid.setTo(-1, invalid!=0); //assert: invalid.channels()==1
  invalid = invalid + 1;
  invalid.convertTo(invalid, CV_8U);

  phase(px, py, pd);
  gdouble pi = atan(1)*4;
  Mat regions;
  pd.copyTo(regions);

  regions.setTo(50, invalid);
  regions.setTo(10, regions < 0.25 * pi);
  regions.setTo(20, regions < 0.5 * pi);
  regions.setTo(40, regions < 0.75 * pi);
  regions.setTo(30, regions < pi);
  regions.setTo(10, regions < 1.25 * pi);
  regions.setTo(20, regions < 1.5 * pi);
  regions.setTo(40, regions < 1.75 * pi);
  regions.setTo(30, regions < 2 * pi);
  regions.setTo(10, regions < 10);

  Mat maskH1, maskH2, maskH3, maskH4;
  regions.copyTo(maskH1, regions==10);
  regions.copyTo(maskH2, regions==20);
  regions.copyTo(maskH3, regions==30);
  regions.copyTo(maskH4, regions==40);
  maskH1.convertTo(maskH1, CV_8U);
  maskH2.convertTo(maskH2, CV_8U);
  maskH3.convertTo(maskH3, CV_8U);
  maskH4.convertTo(maskH4, CV_8U);

  gint histSize = 33334;
  gfloat step = 0.00006;
  gfloat rangeStart = 0.0;
  gfloat rangeEnd = rangeStart + histSize * step;
  gfloat range[] = {rangeStart , rangeEnd};
  const gfloat* histRange = { range };
  Mat hist1, hist2, hist3, hist4;
  calcHist(&pm, 1, 0, maskH1, hist1, 1, &histSize, &histRange, true, false );
  calcHist(&pm, 1, 0, maskH2, hist2, 1, &histSize, &histRange, true, false );
  calcHist(&pm, 1, 0, maskH3, hist3, 1, &histSize, &histRange, true, false );
  calcHist(&pm, 1, 0, maskH4, hist4, 1, &histSize, &histRange, true, false );

  gfloat cumsumH1 = 0;
  gfloat cumsumH2 = 0;
  gfloat cumsumH3 = 0;
  gfloat cumsumH4 = 0;
  gfloat minH;
  for(gint i = histSize-1; i > -1 ; i--){
      cumsumH1 += hist1.at<gfloat>(i,0);
      cumsumH2 += hist2.at<gfloat>(i,0);
      cumsumH3 += hist3.at<gfloat>(i,0);
      cumsumH4 += hist4.at<gfloat>(i,0);
      minH = min(min(cumsumH1, cumsumH2), min(cumsumH3, cumsumH4));
      if(minH >= th){
          mythreshold = rangeStart + i * step;
          break;
      }
  }
  this->threshold = mythreshold;
  adjustThreshold(pm);
  return this->threshold;
}

/**
 * sets the threshold attribute.
 *
 * @param threshold	value to which the threshold needs to be set.
 */
void Kernel::setThreshold(gdouble threshold){
  this->threshold = threshold;
}

/**
 * Resizes this Kernel's image up to the given target size.
 * The resizing is a two-step process. In the first step the
 * kernel is resized by scaling factor sqrt(2). In the second
 * step the size of the kernel is adjusted to match the target
 * size by adding or removing rows and columns.
 * The resizing uses bicubic interpolation based on a cubic
 * Hermite spline.
 * After resizing the kernel is normalized.
 * @param size  the target size of this Kernel's image
 * @throws domain_error  if size is larger than the final
 *                       kernel size
 * @throws domain error  if size is smaller than the current
 *                       current kernel size
 * @throws domain_error  if this Kernel's image is empty
 * @throws domain_error  if this Kernel's image channels > 1
 * @throws domain_error  if this Kernel's image depth < CV_64F
 * @throws domain_error  if the maximum value of this Kernel's image <= 0
 */
void Kernel::resizeForStep(Size size) throw (domain_error)
{
  //refactoring: scale is hard coded
  //refactoring: copy image niet nodig?

  string methodId = "Kernel.resizeForStep";
  string message;
  if(size.width > (gint) this->size || size.height > (gint) this->size){
      message = "size must not be larger than the final kernel size";
      throw domain_error(methodId + ": " + message);
  }
  if(size.width < this->image.cols || size.height < this->image.rows){
      message = "size must not be smaller than the current kernel size";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.empty()){
      message = "this Kernel's image must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.channels() > 1){
      message = "This Kernel's image must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.depth() < CV_64F){
      message = "This Kernel's image depth must be CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  gdouble minVal, maxVal;
  minMaxIdx(this->image, &minVal, &maxVal);
  if(maxVal <= 0){
      message = "This Kernel's image must contain at least one positive value";
      throw domain_error(methodId + ": " + message);
  }

  Mat temp;
  gdouble scale = sqrt(2.0);
  this->image.copyTo(temp);
  temp = resizeImage(temp, scale);
  temp.setTo(0, temp < 0);
  temp = fixSize(temp, size);
  this->image = temp;
  this->normalize();
}

/**
 * Calculates this Kernel's image from the gradients of both
 * the latent image and the blurred image.
 * @param blurred  the blurred image
 * @param latent  the latent image
 * @param size  the expected size of this Kernel's image
 * @throws domain_error  if blurred's image channels > 1
 * @throws domain_error  if blurred's image contains all equal values
 * @throws domain_error  if latent's image channels > 1
 * @throws domain_error  if latent'ś image contains all equal values
 * @throws domain_error  if latent's image size <> blurred's image size
 * @throws domain_error  if size.width = 0 or size.height = 0
 * @throws domain_error  if size.width >= blurred's image cols or size.height >= blurred's image rows
 * @throws domain_error  if this Kernel's image depth < CV_64F
 * @throws domain_error  if this Kernel's image channels > 1
 */
void Kernel::calculateKernel(ImageParameter& blurred, ImageParameter& latent, Size size) throw (domain_error)
{
  string methodId = "Kernel.calculateKernel";
  string message;
  gdouble minVal, maxVal;

  if(blurred.getImage().channels() > 1){
      message = "blurred's image must have no more than one channel";
      throw domain_error(methodId + ": " + message);
  }
  minMaxIdx(blurred.getImage(), &minVal, &maxVal);
  if(minVal==maxVal){
      message = "blurred's image must not have all equal values";
      throw domain_error(methodId + ": " + message);
  }
  if(latent.getImage().channels() > 1){
      message = "latent's image must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }
  minMaxIdx(latent.getImage(), &minVal, &maxVal);
  if(minVal==maxVal){
      message = "latent's image must not have all equal values";
      throw domain_error(methodId + ": " + message);
  }
  if(blurred.getImage().cols != latent.getImage().cols || blurred.getImage().rows != latent.getImage().rows){
    message = "blurred's image and latent's image must have the same size";
    throw domain_error(methodId + ": " + message);
  }
  if(size.width == 0 || size.height == 0){
      message = "size must not be zero";
      throw domain_error(methodId + ": " + message);
  }
  if(size.width >= blurred.getImage().cols || size.height >= blurred.getImage().rows){
      message = "size must be smaller than blurred's image size";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.depth() < CV_64F){
      message = "This Kernel's image depth must be CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.channels() > 1){
      message = "This Kernel's image must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }

  //thresholding
  Mat latent_x = latent.getXGrad();
  Mat latent_y = latent.getYGrad();
  Mat pm = latent.getMagnitude();
  adjustThreshold(pm);
  Mat mask;
  pm.copyTo(mask);
  mask.setTo(this->threshold, mask >= this->threshold);
  mask.setTo(this->threshold + 1, mask < this->threshold);
  mask = mask - this->threshold;
  mask.convertTo(mask,CV_8U);
  latent_x.setTo(0, mask);
  latent_y.setTo(0, mask);
  this->threshold = this->threshold/1.1;

  Mat blurred_xf = fft2(blurred.getXGrad());  //complex Mat
  Mat blurred_yf = fft2(blurred.getYGrad());  //complex Mat
  Mat latent_xf = fft2(latent_x);  //complex Mat
  Mat latent_yf = fft2(latent_y);  //complex Mat
  Mat latent_xfconj = conj(latent_xf);  //complex Mat
  Mat latent_yfconj = conj(latent_yf);  //complex Mat

  Mat mul1, mul2;
  mulSpectrums(latent_xfconj, blurred_xf, mul1, 0, false);
  mulSpectrums(latent_yfconj, blurred_yf, mul2, 0, false);
  Mat b_f = mul1 + mul2;  //complex Mat
  Mat b = otf2psf(b_f, size);

  mulSpectrums(latent_xfconj, latent_xf, mul1, 0, false);
  mulSpectrums(latent_yfconj, latent_yf, mul2, 0, false);
  Mat m_f = mul1 + mul2;  //complex Mat

  Mat psf = Mat::ones(size, this->image.type())/(size.width * size.height); //initial psf

  //conjugate gradient optimization
  Mat ax = computeAx(psf, m_f);
  Mat r = b - ax;
  Mat p, ap;
  r.copyTo(p);
  gint maxIt = 20;
  gdouble tol = 0.00001;
  gdouble rsold = sum(r.mul(r)).val[0];
  gdouble rsnew, alpha;
  gint i;
  for (i = 0; i < maxIt; i++){
      ap = computeAx(p, m_f);
      alpha = rsold / sum(p.mul(ap)).val[0];
      psf = psf + (alpha * p);
      r = r - ( alpha * ap);
      rsnew = sum(r.mul(r)).val[0];
      if(sqrt(rsnew) < tol) break;
      p = r + ((rsnew/rsold) * p);
      rsold = rsnew;
  }

  this->image = psf;
  this->normalize(20);
  this->pruneKernel();
}

/**
 * Sets all pixels of this Kernel's image that have a value
 * below 1/threshold of the maximum value to 0, then divides
 * each pixel's value by the sum of all values.
 * After normalizing the sum of all values becomes 1.
 * If threshold = 0 then all pixels that have a value below 0
 * are set to 0.
 * @param threshold  the threshold for normalizing
 * @throws domain_error  if this Kernel's image is empty
 * @throws domain_error  if the maximum value of this Kernel's image <= 0
 */
void Kernel::normalize(guint threshold) throw (domain_error)
{
  string methodId = "Kernel.normalize";
  string message;
  if(this->image.empty()){
      message = "This Kernel's image must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  gdouble minVal, maxVal;
  Mat temp;
  this->image.copyTo(temp);
  minMaxIdx(temp, &minVal, &maxVal);
  if(maxVal <= 0){
      message = "This Kernel's image must contain at least one positive value";
      throw domain_error(methodId + ": " + message);
  }
  gdouble t = (threshold <= 0) ? 0 : maxVal/threshold;
  temp.setTo(Scalar(0), temp < t);
  gdouble sum1 = sum(temp).val[0];
  temp = (sum1 != 0) ? temp = temp/sum1 : temp;
  this->setImage(temp);
}

/**
 * Sets all pixels of this Kernel's image that have a
 * negative value to 0, then divides each pixel's value
 * by the sum of all values.
 * After normalizing the sum of all values becomes 1.
 * @throws domain_error  if this Kernel's image is empty
 * @throws domain_error  if the maximum value of this Kernel's image <= 0
 */
void Kernel::normalize(void) throw (domain_error)
{
  normalize(0);
}

/**
 * Removes secondary components from this Kernel's image by
 * setting their pixel values to zero.
 * A component is concerned secondary if the sum of values of
 * its pixels is smaller than 0.1.
 * The grouping of pixels into components is based on
 * 8-connected neighborhood.
 * @throws domain_error  if this Kernel's image has no elements
 * @throws domain_error  if this Kernel's image channels > 1
 * @throws domain_error  if this Kernel's depth < CV_64F
 * @throws domain_error  if the maximum value of this Kernel's image <= 0
 */
void Kernel::pruneKernel(void) throw (domain_error)
{
  string methodId = "Kernel.pruneKernel";
  string message;
  if(this->image.total()==0){
      message = "This Kernel's image must not have zero elements";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.depth() < CV_64F){
      message = "This Kernel's image depth must be CV_64F";
      throw domain_error(methodId + ": " + message);
  }

  guint i, j;
  gdouble threshold = 0.1;
  gdouble sum1;
  Mat labeled;
  vector<vector<Point> > regions;
  vector<vector<Point> > noise;

  connectedComponents(this->image, labeled, regions);

  // identify noise
  for (i = 0; i < regions.size(); i++){
      sum1 = 0.0;
      j = 0;
      for ( j = 0; j < regions[i].size() ; j++){
          sum1 += this->image.at<gdouble>(regions[i][j]); //CV_64F
      }
      if(sum1 < threshold) {
          noise.push_back(regions[i]);
      }
  }

  // remove noisy regions
  for (i = 0; i < noise.size(); i++){
      for (j = 0; j < noise[i].size() ; j++){
          this->image.at<gdouble>(noise[i][j]) = 0.0;  //CV_64F
      }
  }

  this->normalize();
}


/**
 * Helper method for calculating the kernel.
 * Calculates the Ax-part of the formula r = B - Ax
 * used for conjugate gradient optimization.
 * @param x  intermediate kernel
 * @param p  complex matrix calculated from the gradient
 *           maps of the latent and the blurred image
 * @return  the Ax-part of the formula r = B - Ax
 *          used for conjugate gradient optimization.
 * @throws domain_error  if x is empty
 * @throws domain_error  if x.depth < CV_64F
 * @throws domain_error  if x.channels <> 1
 * @throws domain_error  if p.cols < x.cols or p.rows < x.rows
 * @throws domain_error  if p.depth < CV_64F
 * @throws domain_error  if p.channels <> 2
 */
Mat Kernel::computeAx(const Mat& x, const Mat& p) throw (domain_error)
{

  //refactoring: convert x.depth, p.depth to CV_64F

  string methodId = "Kernel.computeAx";
  string message;
  if(x.empty()){
      message = "x must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  if(x.depth() < CV_64F){
      message = "x depth must be CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  if(x.channels() > 1){
      message = "x must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }
  if(p.cols < x.cols || p.rows < x.rows){
      message = "p size must not be smaller than x size";
      throw domain_error(methodId + ": " + message);
  }
  if(p.depth() < CV_64F){
      message = "p depth must be CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  if(p.channels() != 2){
      message = "p must have exactly two channels";
      throw domain_error(methodId + ": " + message);
  }

  Mat xf = psf2otf(x, Size(p.cols, p.rows));  //complex Mat
  Mat mul; //complex Mat
  Mat y;
  mulSpectrums(p, xf, mul, 0, false);
  y = otf2psf(mul, Size(x.cols, x.rows));
  y = y + (2.0 * x);
  return y;
}

/**
 * Shifts this Kernel's image until the center of mass reaches
 * the central position.
 * Empty rows and columns as a result of the shift are
 * zero-valued.
 * After shifting the kernel is normalized.
 * throws domain_error  if this Kernel's image is empty
 * throws domain_error  if this Kernel's image width or height is even
 * throws domain_error  if this Kernel's image channels > 1
 * throws domain_error  if this Kernel's image depth < CV_64F
 * throws domain_error  if the maximum value of this Kernel's image <= 0
 */
void Kernel::adjustPSFCenter(void) throw (domain_error)
{
  string methodId = "Kernel.adjustPSFCenter";
  string message;
  if(this->image.empty()){
      message = "this Kernel's image must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.cols % 2 == 0 || this->image.rows % 2 == 0){
      message = "both this Kernel's image width and height must be odd";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.channels() > 1){
      message = "This Kernel's image must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }
  if(this->image.depth() < CV_64F){
      message = "This Kernel's image depth must be CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  gdouble minVal, maxVal;
  minMaxIdx(this->image, &minVal, &maxVal);
  if(maxVal <= 0){
      message = "This Kernel's image must contain at least one positive value";
      throw domain_error(methodId + ": " + message);
  }

  Mat Ix, Iy, Ix64F, Iy64F, psf_adjusted;
  gdouble xc1, xc2, yc1, yc2;
  gint xshift, yshift;
  meshgrid(Size(this->image.cols, this->image.rows), Ix, Iy, 1, 1);
  Ix.convertTo(Ix64F, CV_64FC1);
  Iy.convertTo(Iy64F, CV_64FC1);
  xc1 = sum(this->image.mul(Ix64F)).val[0];
  xc2 = (this->image.cols + 1)/2.0;
  yc1 = sum(this->image.mul(Iy64F)).val[0];
  yc2 = (this->image.rows + 1)/2.0;
  xshift = round(xc2 - xc1);
  yshift = round(yc2 - yc1);
  meshgrid(Size(this->image.cols, this->image.rows), Ix, Iy, -xshift, -yshift);
  psf_adjusted = remapImage(this->image, Ix, Iy);
  this->image = psf_adjusted;
  this->normalize();
}

/**
 * Creates two x- and y- coordinates maps with the given size.
 * The coordinates are shifted by the given xshift and yshift.
 * The output coordinates maps have type CV_32SC1.
 * The output coordinates maps can be used, for example, for
 * remapping an image.
 * @param size  the size of the output coordinates maps
 * @param x  output parameter for the x-coordinates map
 * @param y  output parameter for the y-coordinates map
 * @param xshift  number of positions the x-coordinates are shifted
 * @param yshift  number of positions the y-coordinates are shifted
 * @throws domain_error  if size width < 1 or size height < 1
 */
void Kernel::meshgrid(Size size, Mat& x, Mat& y, gint xshift, gint yshift) throw (domain_error)
{
  //refactoring: volgorde in- en outparameters

  string methodId = "Kernel.meshgrid";
  string message;
  if(size.width < 1 || size.height < 1){
      message = "size must be at least (1,1)";
      throw domain_error(methodId + ": " + message);
  }

  x = Mat(size, CV_32SC1);
  y = Mat(size, CV_32SC1);
  gint i;
  for(i = 0; i< size.height; i++){
      y.row(i).setTo(i + yshift);
  }
  for(i = 0; i< size.width; i++){
      x.col(i).setTo(i + xshift);
  }
}

/**
 * Remaps the elements of the given matrix according to the
 * size and values of the given coordinate maps.
 * The values of non-integer coordinates are calculated using
 * bilinear interpolation. The values of elements that
 * correspond to outliers in the input matrix are set to zero.
 * @param src  the input matrix
 * @param Ix  map containing the x-coordinates for remapping
 * @param Iy  map containing the y-coordinates for remapping
 * @return  the remapped matrix
 * @throws domain_error  if src has no elements
 * @throws domain_error  if Ix has no elements or Iy has no elements
 * @throws domain_error  if Ix and Ix have different sizes
 * @throws domain_error  if Ix depth > CV_32F or Iy depth > CV_32F
 * @throws domain_error  if Ix channels > 1 or Iy channels > 1
 */
Mat Kernel::remapImage(const Mat& src, Mat& Ix, Mat Iy) throw (domain_error)
{
  string methodId = "Kernel.remapImage";
  string message;
  if(src.total()==0){
      message = "src must not have zero elements";
      throw domain_error(methodId + ": " + message);
  }
  if(Ix.total()==0 || Iy.total()==0){
      message = "Ix and Iy must not have zero elements";
      throw domain_error(methodId + ": " + message);
  }
  if(Ix.size != Iy.size){
      message = "Ix and Iy must have the same size";
      throw domain_error(methodId + ": " + message);
  }
  if(Ix.depth() > CV_32F ||  Iy.depth() > CV_32F){
      message = "Ix depth and Iy depth must be at most CV_32F";
      throw domain_error(methodId + ": " + message);
  }
  if(Ix.channels() > 1 || Iy.channels() > 1){
      message = "Ix and Iy must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }

  Mat dst, Ix32FC1, Iy32FC1;
  Ix.convertTo(Ix32FC1, CV_32FC1);
  Iy.convertTo(Iy32FC1, CV_32FC1);
  remap(src, dst, Ix32FC1, Iy32FC1, CV_INTER_LINEAR);
  return dst;
}


/**
 * Changes the number of rows of the given matrix by rearranging its
 * elements.
 * The reshaped matrix maintains the same number of elements as the original
 * matrix. The elements of the matrix are processed in column-wise order.
 * @param src  the input matrix
 * @param rows  the number of rows of the reshaped matrix
 * @return  the reshaped matrix having the given number of rows
 * @throws domain_error  if rows < 1
 * @throws domain_error  if rows does not divide evenly into the number
 *                       of elements of src
 */
Mat Kernel::reshapeImage(const Mat& src, gint rows) throw (domain_error)
{
  //refactoring: if rows == src.rows return a copy of src
  //refactoring: guint rows

  string methodId = "Kernel.reshapeImage";
  string message;
  if(rows < 1){
      message = "rows must be at least 1";
      throw domain_error(methodId + ": " + message);
  }
  if(src.total() % rows != 0){
      message = "rows must divide evenly into the number of elements of src";
      throw domain_error(methodId + ": " + message);
  }

  gint i;
  gint cols = src.total()/rows;
  Mat reshaped = Mat(rows, cols, src.type());
  Mat src_tp;
  transpose(src, src_tp);
  Mat src_tp_rs = src_tp.reshape(0, src.total());
  for (i = 0; i < cols; i++){
      src_tp_rs(Rect(0, i*rows, 1, rows)).copyTo(reshaped(Rect(i, 0, 1, rows)));
  }
  return reshaped;
}


/**
 * Helper method for pruning the kernel.
 * Computes the connected components of src based on
 * 8-connected neighborhood.
 * @param src  the input matrix
 * @param labeled  outparameter for a matrix containing the labeled components
 * @param regions  outparameter for a vector containing the coordinates of the pixels of each component
 * @throws domain_error  if src is empty
 * @throws domain_error  if src channels > 1
 */
void Kernel::connectedComponents(const Mat& src, Mat& labeled, vector<vector<Point> >& regions) throw (domain_error)
{
  string methodId = "Kernel.connectedComponents";
  string message;
  if(src.empty()){
      message = "src must not be empty";
      throw domain_error(methodId + ": " + message);
  }
  if(src.channels() > 1){
      message = "src must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }

  regions.clear();

  gint x, y, i, j;
  Rect rect;
  gint label = 2;
  src.copyTo(labeled);
  labeled.setTo(cv::Scalar(1), labeled != 0);
  //cout << "binary = " << endl << " " << labeled << endl << endl;
  labeled.convertTo(labeled, CV_32SC1);
  for ( y = 0 ; y < labeled.cols ; y++){
    for ( x = 0; x < labeled.rows ; x++){
      if (labeled.at<int>(x, y) == 1) {
        floodFill(labeled, cv::Point(y, x), cv::Scalar(label), &rect, 0, 0, 8);
        vector<Point> region;
        for( i = rect.y; i < (rect.y + rect.height); i++) {
          for( j = rect.x; j < (rect.x + rect.width); j++) {
            int reg = labeled.at<gint>(i,j);
            if(reg == label) {
              region.push_back(Point(j,i));
            }
          }
        }
        regions.push_back(region);
        label++;
      }
    }
  }
}


/**
 * Resizes the given image by the given scaling factor.
 * The resizing uses bicubic interpolation based on a cubic
 * Hermite spline.
 * @param src  the input image
 * @param scale  the scaling factor
 * @return  the resized image
 * @throws domain_error  if src has no elements
 * @throws domain_error  if src channels > 1
 * @throws domain_error  if src depth < CV_64F
 * @throws domain_error  if scale < 0.00001
 * @throws domain_error  if scale * src.cols > 10.000 or scale * src.rows > 10.000
 */
Mat Kernel::resizeImage(const Mat& src, gdouble scale) throw (domain_error)
{
  string methodId = "Kernel.resizeImage";
  string message;

  if(src.total() == 0){
      message = "src must not have zero elements";
      throw domain_error(methodId + ": " + message);
  }
  if(src.channels() > 1){
      message = "src must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }
  if(src.depth() < CV_64F){
      message = "src depth must be CV_64F";
      throw domain_error(methodId + ": " + message);
  }
  if(scale < 0.00001){
      message = "scale must be at least 0.00001";
      throw domain_error(methodId + ": " + message);
  }
  if(scale * src.cols > 10000 || scale * src.rows > 10000 ){
      message = "output size must not exceed 10.000 x 10.000";
      throw domain_error(methodId + ": " + message);
  }

  Mat rescaled;
  Mat temp;
  src.copyTo(temp);
  if(scale == 1){
      src.copyTo(rescaled);
  }
  else{
      Mat padded;
      gint xsize;
      gint i, j, k;
      gint minxcoord, xcoord;
      gdouble xloc, distance;
      gdouble weights[4];
      gdouble newval;
      for (gint dim = 1; dim < 3; dim++){
          copyMakeBorder(temp, padded, 0, 0, 2, 2, BORDER_REFLECT);
          xsize = ceil(scale * temp.cols);
          rescaled = Mat::zeros(temp.rows, xsize, temp.type());
          for(i = 0; i < xsize; i++){
              xloc = (i + 0.5) * (1/scale);
              if(xloc - floor(xloc) < 0.5){
                  minxcoord = floor(xloc) - 2;
              }
              else{
                  minxcoord = floor(xloc) - 1;
              }
              distance = xloc - minxcoord - 0.5;
              weights[0] = cubicWeight(distance);
              weights[1] = cubicWeight(distance-1);
              weights[2] = cubicWeight(distance-2);
              weights[3] = cubicWeight(distance-3);
              for (j = 0; j < temp.rows; j++){
                  newval = 0;
                  for (k = 0; k < 4; k++){
                      xcoord = minxcoord + k + 2;
                      newval += padded.at<gdouble>(j, xcoord) * weights[k];
                  }
                  rescaled.at<gdouble>(j,i) = newval;
              }
          }
          transpose(rescaled, temp);
      }
      rescaled = temp;
  }
  return rescaled;
}


/**
 * Adjusts the size of the given matrix to match the
 * target size. If the width of the target size is
 * larger than the width of the input matrix then colums
 * are added. Conversely, if the width of the target size
 * is smaller than the width of input matrix then columns
 * are removed.
 * The height of the input matrix is adjusted the same way.
 * Adjustments are located at the borders of the
 * matrix, in such way that the original values of the
 * matrix are maintained as much as possible.
 * Added rows and columns are zero-valued.
 * @param src  the input matrix
 * @param size  the target size
 * @return  adjusted matrix having the given size
 * @throws domain_error  if src channels > 1
 * @throws domain_error  if src has no elements
 * @throws domain_error  if size.width < 1 or size.height < 1
 */
Mat Kernel::fixSize (const Mat& src, Size size) throw (domain_error)
{
  string methodId = "Kernel.fixSize";
  string message;
  if(src.channels() > 1){
      message = "src must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }
  if(src.total() == 0){
      message = "src must not have zero elements";
      throw domain_error(methodId + ": " + message);
  }
  if(size.width < 1 || size.height < 1 ){
      message = "size must be at least (1,1)";
      throw domain_error(methodId + ": " + message);
  }

  gdouble sumfirstrow, sumlastrow, sumfirstcol, sumlastcol;
  Mat temp;
  src.copyTo(temp);

  while (temp.rows != size.height || temp.cols != size.width){

      sumfirstrow = sum(temp(Rect(0,0,temp.cols, 1)))[0];
      sumlastrow = sum(temp(Rect(0,temp.rows-1,temp.cols, 1)))[0];

      if(temp.rows > size.height) {
          //cout << "Case 1: rows > height" << endl;
          if (sumfirstrow < sumlastrow){
              temp = temp(Rect(0,1, temp.cols, temp.rows-1));
          }
          else{
              temp = temp(Rect(0,0, temp.cols, temp.rows-1));
          }
      }
      else if(temp.rows < size.height) {
          //cout << "Case 2: rows < height" << endl;
          if (sumfirstrow < sumlastrow){
              copyMakeBorder(temp, temp, 0, 1, 0, 0, BORDER_CONSTANT, 0 );
          }
          else{
              copyMakeBorder(temp, temp, 1, 0, 0, 0, BORDER_CONSTANT, 0 );
          }
      }

      sumfirstcol = sum(temp(Rect(0,0,1, temp.rows)))[0];
      sumlastcol = sum(temp(Rect(temp.cols-1,0,1, temp.rows)))[0];

      if(temp.cols > size.width) {
          //cout << "Case 3: cols > width" << endl;
          if (sumfirstcol < sumlastcol){
              temp = temp(Rect(1,0, temp.cols-1, temp.rows));
          }
          else{
              temp = temp(Rect(0,0, temp.cols-1, temp.rows));
          }
      }
      else if(temp.cols < size.width) {
          //cout << "Case 4: cols < width" << endl;
          if (sumfirstcol < sumlastcol){
              copyMakeBorder(temp, temp, 0, 0, 0, 1, BORDER_CONSTANT, 0 );
          }
          else{
              copyMakeBorder(temp, temp, 0, 0, 1, 0, BORDER_CONSTANT, 0 );
          }
      }
  }
  return temp;
}

/**
 * Helper method for the calculation of cubic interpolation.
 * Calculates the cubic weight for the sample point at the given
 * distance to the inquiry point. The calculation of the
 * weight is based on a cubic Hermite spline.
 * @param distance  the distance from the sample point
 *                  to the inquiry point
 * @return  the cubic weight for the sample point at the given
 *          distance to the inquiry point
 */
gdouble Kernel::cubicWeight(const gdouble distance)
{
  gdouble d = abs(distance);
  gdouble weight = 0;
  gdouble a = -0.50; //de boosdoener
  if (d==0){
      weight = 1;
  }
  else if (d==1 || d>=2){
      weight = 0;
  }
  else {
      gdouble d2 = pow(d, 2.0);
      gdouble d3 = pow(d, 3.0);
      if(d<1){
          weight = (a+2)*d3 - (a+3)*d2 + 1;
      }
      else{
          weight = a*d3 - 5*a*d2 + 8*a*d - 4*a;
      }
  }
  return weight;
}

/**
 * Adjusts this Kernel's threshold for discarding the gradients
 * of the latent image during kernel calculation.
 * The threshold is calculated from the squared magnitudes of
 * the latent image. If the latent image is not yet calculated
 * the blurred image is used instead.
 * @param magnitude  map containing the squared magnitudes of the latent image
 * @throws domain_error  if magnitude channels > 1
 * @throws domain_error  if magnitude contains any negative values
 * @throws domain_error  if magnitude's maximum value == 0 and this Kernel's threshold > 0
 */
void Kernel::adjustThreshold(const Mat& magnitude) throw (domain_error)
{
  string methodId = "Kernel.adjustThreshold";
  string message;

  if(magnitude.channels() > 1){
      message = "magnitude must have exactly one channel";
      throw domain_error(methodId + ": " + message);
  }

  gdouble minVal, maxVal;
  gdouble mythreshold = this->threshold;
  minMaxIdx(magnitude, &minVal, &maxVal); // assert: magnitude.channels() == 1

  if(minVal < 0){
      message = "magnitude must not contain any negative values";
      throw domain_error(methodId + ": " + message);
  }
  if(maxVal == 0 && this->threshold > 0){
      message = "magnitude's maximum value must be > 0 if this Kernel's threshold > 0";
      throw domain_error(methodId + ": " + message);
  }

  while(maxVal < mythreshold){
      mythreshold = mythreshold * 0.81;
  }
  this->threshold = mythreshold;
}

