/*
 * Class Kernel represents the blur kernel.
 *
 * @author  E. van Schaijk
 * @author  T. Poppe
 * @version  1.0
 * @since  July, 2016
 *
 * This code implements the image deblurring algorithm described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 */

#ifndef TEXTDEBLURPAN_KERNEL_HPP_
#define TEXTDEBLURPAN_KERNEL_HPP_

#include <opencv2/core/core.hpp>
#include <stdexcept>

#include "ImageParameter.hpp" // Include header of the base class

/**
 * class representing the kernel
 */
class Kernel : public ImageParameter {

private:

  /*
   * The final size of this Kernel's image.
   */
  guint size;

  /*
   * The upper bound for discarding the gradients of
   * the latent image during kernel calculation.
   */
  gdouble threshold;

  Kernel(void); // private default constructor
  static gdouble cubicWeight(const gdouble distance);


public:
  Kernel(guint size) throw (std::domain_error);
  guint getSize(void);
  void initialize(guint size) throw (std::domain_error);
  void resizeForStep(cv::Size size) throw (std::domain_error);
  void normalize(guint threshold) throw (std::domain_error);
  void calculateKernel(ImageParameter& blurred, ImageParameter& latent, cv::Size size) throw (std::domain_error);
  void adjustPSFCenter(void) throw (std::domain_error);
  gdouble initializeThreshold(ImageParameter& blurred) throw (std::domain_error);
  void setThreshold(gdouble threshold); //voor testdoeleinden

  //nog private maken
  void pruneKernel(void) throw (std::domain_error);
  void normalize(void) throw (std::domain_error);
  static void meshgrid(cv::Size size, cv::Mat& x, cv::Mat& y, gint xshift, gint yshift) throw (std::domain_error);
  static cv::Mat remapImage(const cv::Mat& src, cv::Mat& Ix, cv::Mat Iy) throw (std::domain_error);
  static cv::Mat reshapeImage(const cv::Mat& src, gint rows) throw (std::domain_error);
  static void connectedComponents(const cv::Mat& src, cv::Mat& labeled, std::vector<std::vector<cv::Point> >& regions) throw (std::domain_error);
  static cv::Mat resizeImage(const cv::Mat& src, gdouble scale) throw (std::domain_error);
  static cv::Mat fixSize (const cv::Mat& src, cv::Size size) throw (std::domain_error);
  static cv::Mat computeAx(const cv::Mat& x, const cv::Mat& p) throw (std::domain_error);
  void adjustThreshold(const cv::Mat& magnitude) throw (std::domain_error);


};

#endif /* TEXTDEBLURPAN_KERNEL_HPP_ */
