/*
 * LatentImage.cpp
 *
 * @author  E. van Schaijk
 * @author  T. Poppe
 * @version  1.0
 * @since  July, 2016
 *
 * Part of the implementation of the image deblurring algorithm described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 *
 * EXCLUDE =  FrameworkControllerBatch.hpp
 *
 */

#include "LatentImage.hpp"

#include <opencv2/core/core.hpp>
#include "opencv2/imgproc/imgproc.hpp"
#include <opencv2/highgui/highgui.hpp>
#include <iostream>
#include <stdexcept>
#include "BlurredImage.hpp"

using namespace cv;
using namespace std;

/**
 * Create the LatentImage object
 */
LatentImage::LatentImage() {}

/**
 * Initialized the latentImage object by creating an empty matrix with the size given as input parameter.
 * The empty matrix coordinates are initialized to zero.
 *
 * @param size 		array with the number of rows and columns to which the initial latent image needs to be initialized.
 * @throw domain_error  if size is smaller then 1.
 */
void LatentImage::initialize(gint size[2]) throw (domain_error){
  // validate size which should be at least 1x1
  if (size[0] < 1 || size[1] < 1){
	  throw domain_error("size is smaller then 1");
  }
  // Todo is misschien overbodig ?
  this->image = Mat::zeros(size[1], size[0], CV_64FC1);

}


/**
 * Calculates the latent image based on the given kernel and blurred image.
 * The latent image calculation is based on the algorithm of Pan et al.
 *
 * @param kernel	ImageParameter representing the kernel.
 * @param blurred	BlurredImage representing the blurred image.
 * @param lambdaPixel parameter determining the weight of the L0 pixel intensity prior.
 * @param lambdaGrad  parameter determining the weight of the L0 gradient prior.
 *
 * @throw domain_error if either lambdaPixel or lambdaGrad <0
 */
void LatentImage::calculateLatent(ImageParameter& kernel, BlurredImage& blurred,
    gdouble lambdaPixel, gdouble lambdaGrad) throw (domain_error){

  if(lambdaPixel <0){
	throw domain_error("LatentImage - calculateLatent: lambdaPixel can't be negative ");
  }
  if (lambdaGrad <= 0 ){
    throw domain_error("LatentImage - calculateLatent: lambdaGrad can't be zero ");
  }

  blurred.setImagePadding();
  this->image = blurred.getImage().clone();
  int channels = image.channels();
  double muMax = 1E5;
  double betaMax = pow(2,3);
  Size size = blurred.getImage().size();

  // Kfft ← psf2otf(k, size)
  Mat Kfft = ImageParameter::psf2otf(kernel.getImage(), size); // ok

  //denomin2 ← abs(psf2otf(fx,size))²+abs(psf2otf(fy,size))²
  Mat denormin2 = calculateDenomin2(size);

  //nomin1 ← conj(Kfft).*fft(b)
  Mat b = blurred.getImage();
  Mat normin1[4];
  Mat bDim[4];
  split(b,bDim);
  for (int ch=0; ch<channels ; ch++){
	normin1[ch] = calculateNomin1(Kfft,bDim[ch]);
  }

  //den_ker ← abs(Kfft)²
  Mat denKer = ImageParameter::absModules(Kfft);
  pow(denKer, 2, denKer);

  //beta ← 2 * lp
  double beta = 2*lambdaPixel;

  while (beta < betaMax){

	this->solveU(lambdaPixel , beta);

	double mu = 2*lambdaGrad;
	while (mu < muMax){

	  this->solveG(lambdaGrad, mu);
	  Mat dg = this->gradientG();

	  //fg ← fft(dg)
	  Mat fg[4];
	  Mat dgDim[4];
	  split(dg,dgDim);
	  for (int ch=0; ch<channels ; ch++){
		  fg[ch]= ImageParameter::fft2(dgDim[ch]);
	  }

	  //Denom ← den_ker + mu * denomin2 + beta
	  Mat denorm = denKer + denormin2 * mu + beta; //ok

	  //sfft ← normin1 + mu*fg + beta*fft(u) / denomrin
	  Mat sfft[4];
	  Mat sfft2[4];
	  Mat res[4];
	  Mat uDim[4];
	  split(u,uDim);
	  for (int ch=0; ch<channels ; ch++){
		  sfft[ch] = normin1[ch] + mu*fg[ch] + beta*ImageParameter::fft2(uDim[ch]);
		  sfft2[ch] = ImageParameter::divComplex(sfft[ch], denorm);
		  //l ← real(ifft(fs))
		  res[ch] = ImageParameter::ifft2(sfft2[ch]);
	  }

	  merge(res,channels, image);

	  mu = mu*2;

	}

	// check if lambdaPixel is null and there need to skip the beta loop
	if (lambdaPixel == 0){
	  beta = betaMax;
	} else {
	  beta= beta*2;
	}


  }

  // remove padding
  this->cropImage(blurred.getSizeRescaled());
  blurred.setImageNoPadding();
}

/**
 * Calculates the support variable G representing the L0 gradient prior
 * refer to algorithm Pan et al for details.
 *
 * @param lambdaGrad parameter determining the weight of the L0 gradient prior.
 * @param mu 		parameter used for determining the cut-off threshold
 * @param applyThreshold  flag to indicate whether the mu cut-off threshold needs to be applied. default values is true.
 *
 * @throw domain_error if no image is set or if mu =0 and applythreshold is true.
 */
void LatentImage::solveG(gdouble lambdaGrad, gdouble mu,bool applyThreshold) throw (domain_error){
  // validatie current image
  if (image.empty()){
	  throw domain_error("No image is set");
  }

  if (image.depth() != CV_64F){
    image.convertTo(image, CV_64FC1);
  }



  // calculate x gradient of image
  Mat firstColumn = image.col(0);
  Mat lastColumn = image.col(image.cols-1);

  Mat h1 = image.colRange(0, image.cols-1);
  Mat h2 = image.colRange(1, image.cols);

  h.release();
  Mat temp = h2 - h1;
  cv::hconcat(temp,firstColumn - lastColumn,h);

  // calculate y gradient of image
  Mat firstRow = image.row(0);
  Mat lastRow = image.row(image.rows-1);

  Mat v1 = image.rowRange(0, image.rows-1);
  Mat v2 = image.rowRange(1, image.rows);

  v.release();
  temp.release();
  temp = v2-v1;
  cv::vconcat(temp, firstRow - lastRow, v);

  // if applyThreshold = true then values with H² + v² < lamdbaGrad/mu are set to zero
  if(applyThreshold){
	if (mu==0){
	  throw domain_error("LatentImage::SolveG: mu can be zero when threshold is applied ");
	}

	//split h en v into channels
    int channels = h.channels();
    Mat planesH[4];
    split(h, planesH);
    Mat planesV[4];
    split(v, planesV);

    Mat vPower2;
    Mat hPower2;
    Mat total;
    // start of loop per channel to calculate the H² + v² summed for all channels
    for (int ch=0 ; ch<channels; ch++){
      // calculate power 2 of h en v and add to total
      pow(planesH[ch],2,hPower2);
      pow(planesV[ch],2,vPower2);
      if(total.empty()){
    	hPower2.copyTo(total);
    	total = total + vPower2;
      } else {
    	total = total + hPower2 + vPower2;
      }
    }


    // create matrix t with 0 or 1 depending if H² + v² < lamdbaGrad/mu
    Mat t = Mat::ones(h.size(), CV_64FC1);
    t.setTo(0, total < lambdaGrad/mu);

    // create multi channel t matrix
    for (int ch=0 ; ch<channels; ch++){
    	planesH[ch] = t;
    }
    merge(planesH, channels, t);

    h=h.mul(t);
    v=v.mul(t);
  }

}

/**
 * calculates the matrix u. u is the supporting matrix representing the L0 pixel intensity prior.
 * refer to algorithm of Pan et al. for more details.
 *
 * @param lambdaPixel parameter determining the weight of the L0 pixel intensity prior.
 * @param beta		parameter used for determining the cut-off threshold
 *
 * @throw domain_error if no image is set or if beta <=0 .
 *
 */
void LatentImage::solveU(gdouble lambdaPixel, gdouble beta) throw (domain_error){

  // validatie current image
  if (image.empty()){
	  throw domain_error("No image is set");
  }

  if (image.depth() != CV_64F){
	image.convertTo(image, CV_64FC1);
  }

  u.release();
  image.copyTo(u);

  if (lambdaPixel != 0){
    if (beta <= 0){
	  throw domain_error("beta can't be zero");
    }


    int channels = u.channels();
    Mat planes[4];
    split(u, planes);
    Mat uPower2;
    Mat uPower2Total;
    Mat t = Mat::ones(u.size(), CV_64FC1);

    for (int ch=0 ; ch<channels; ch++){

	  pow(planes[ch],2, uPower2);
	  if(uPower2Total.empty()){
		uPower2.copyTo(uPower2Total);
	  } else {
	    uPower2Total = uPower2Total + uPower2;
	  }
    }

    t.setTo(0, uPower2Total < lambdaPixel/beta);

	for (int ch=0 ; ch<channels; ch++){
	  planes[ch] = t;
	}
    merge(planes,channels,t);

    u = u.mul(t);
  }
}

/**
 * calculates the gradient of the support variabele G
 *
 * @throw domain_error if no horizontal of vertical component of G is set.
 */
Mat LatentImage::gradientG() throw (domain_error){
  // check if h and v are set
  if (h.empty() || v.empty()){
	  throw domain_error("h or v is not set");
  }

  // calculate h gradient
  Mat firstColumn = this->h.col(0);
  Mat lastColumn = this->h.col(this->h.cols-1);

  Mat dh1 = this->h.colRange(0, this->h.cols-1);
  Mat dh2 = this->h.colRange(1, this->h.cols);

  Mat dh;
  Mat temp = lastColumn - firstColumn;
  cv::hconcat(temp, dh1-dh2 ,dh);
  temp.release();

  Mat firstRow = this->v.row(0);
  Mat lastRow = this->v.row(this->v.rows-1);

  Mat dv1 = this->v.rowRange(0, this->v.rows-1);
  Mat dv2 = this->v.rowRange(1, this->v.rows);

  Mat dv;
  temp = lastRow - firstRow;
  cv::vconcat(temp, dv1- dv2 , dv);

  return (dh+dv);
}


/**
 * Final deconvolution with a non blind Laplacian deconvolution.
 * Depending on the weight factor for ringing removal, the final deconvolution includes
 * a ringing removal by calculation the difference between LaPlacian and L0 intensity deconvolution.
 * A bilateral filter is applied to this difference and subtraced from the LaPlacian deconvolution.
 *
 * @param kernel	ImageParameter representing the kernel.
 * @param blurred	Blurredimage representing the blurred image.
 * @param lambdaTV	parameter determining the weight of the Laplacian prior.
 * @param lambda	parameter determining the weight of the L0 intensity prior.
 * @param weightRinging factor determining the weight for the ringing removal via bilateral filtering
 * @return a Mat object representing the latent image.
 *
 * @throw domain_error if weightRinging, lambdaTV or lambda L0 is lower then 0.
 */
Mat LatentImage::calculateLaPlacian(ImageParameter kernel, BlurredImage blurred,
		double lambdaTV, double lambda, double weightRinging) throw (domain_error){

  // validate input
  if(weightRinging < 0 || lambdaTV < 0 || lambda < 0){
	throw domain_error("LatentImage::calculateLaplaciain: weightRinging, lambdaTV and LambdaL0 can't be negative.");
  }
  Mat blurredMat = blurred.getImageOriginal().clone();
  cv::Size size = blurredMat.size();
  int channels = blurredMat.channels();

  Mat planes[4];
  Mat latent[4];
  split(blurredMat, planes); // planes[0] = blue, planes[1] = green, planes[2] = red, planes[3] = alpha if existing
  for ( int i =0; i<channels ; i++){
	blurred.setImageOriginalOneChannel(planes[i], false);
	blurred.resizeForFFT(1,kernel.getCurrentHeight());

	deblurringAdmAniso(blurred.getImage(), kernel.getImage(), lambdaTV);
	cropImage(size);
	latent[i] = getImage();
  }
  Mat latentColor;
  merge(latent, channels, latentColor);

  // check if ringing removal needs to be applied
  if (weightRinging != 0){
	 //calculate non blind L0deblur
	 blurred.resizeForFFT( 1, kernel.getCurrentHeight(), false );
	 calculateLatent(kernel, blurred, 0,lambda);
	 cropImage(size);

	 // calculate diffirence between LaPlacian en L0deblur
	 Mat diff;
	 latentColor.copyTo(diff);
	 diff = diff - getImage();
	 Mat bfDiff = this->biLateralFilterPan(diff,3.0, 0.1);

	 // Apply bilateral filter to different and subtract from LaPlacian deblur with factor = weightRinging
	 //cv::bilateralFilter(diff, bfDiff, 0, 0.5, 3.0);
	 bfDiff.convertTo(bfDiff,CV_64F);
	 latentColor = latentColor - weightRinging * bfDiff;
  }

  //remove outliers
  channels =  latentColor.channels();
  Mat latentColorPlanes[channels];
  split(latentColor, latentColorPlanes);
  for(int i = 0; i < channels ; i++){
      latentColorPlanes[i].setTo(0, latentColorPlanes[i] < 0);
      latentColorPlanes[i].setTo(1, latentColorPlanes[i] > 1);
  }
  merge(latentColorPlanes, channels, latentColor);

  latentColor.copyTo(this->image);
  return this->image;
}

/**
 * non blind deconvolution using the Laplacian prior.
 *
 * @param blurred  Mat object representing the blurred image.
 * @param kernel   	Mat object representing the kernel
 * @param lambdaTV	Parameter which determines the weight of the Laplacian prior.
 * @return a mat object representing the latent image.
 * @throw domain_error if lambdaTV <0.
 */
cv::Mat LatentImage::deblurringAdmAniso(cv::Mat blurred, cv::Mat kernel, double lambdaTV) throw (domain_error){

  // validate input
  if(lambdaTV < 0){
	throw domain_error("LatentImage::deblurringAdmAniso: lambdaTV can't be negative");
  }
  this->image = blurred.clone();


  cv::Size size = blurred.size();

  //otfK ← psf2otf(k,size)
  Mat otfK = ImageParameter::psf2otf(kernel, size);

  //nomin1 = conj(otfK).*fft2(b)
  Mat nomin1 = calculateNomin1(otfK,blurred);

  //denom1 ← abs(otfK).^2
  Mat denom1;
  pow(ImageParameter::absModules(otfK),2, denom1);

  //denomin2 ← abs(psf2otf(fx,size))²+abs(psf2otf(fy,size))²
  Mat denomin2 = calculateDenomin2(size);

  //Beta
  double beta = 1/ lambdaTV;
  double beta_min = 0.001;
  while( beta>beta_min){
    //bereken  h en v I
	this->solveG(0,0,false);

	//denom = denom1 + gamma * denomin2
	double gamma = 1/(2*beta);
	Mat denom = denom1 + denomin2*gamma;

	//Wh = max(abs(h) – beta*tv, 0).* sign(h)
	Mat temp = abs(h) - beta*lambdaTV;
	h = max(temp,0).mul(ImageParameter::signum(h));

	//Wv = max(abs(v) – beta*tv,0).* sign(v)
	temp = abs(v) - beta*lambdaTV;
	v = max(temp,0).mul(ImageParameter::signum(v));

    Mat wxx = gradientG();

    //out ← (nomin1 + gamma*fft2(wxx))./Denom
    Mat nomin = nomin1 + gamma*ImageParameter::fft2(wxx);
    Mat out = ImageParameter::divComplex(nomin, denom);
    Mat res = ImageParameter::ifft2(out);

    res.copyTo(this->image);

    beta = beta/2;
  }

  return this->image;
}

/**
 * returns the horizontal gradient component of the support variable G.
 *
 * @return horizontal gradient of support variable G.
 */
Mat LatentImage::getH(){
	return this->h;
}

/**
 * returns the vertical gradient component of the support variable G.
 * G represents the L0 gradient variable.
 *
 * @return vertical gradient of support variable G.
 */
Mat LatentImage::getV(){
	return this->v;
}

/**
 * returns the support variable U representing the L0 pixel intensity.
 *
 * @return support variable u.
 */
Mat LatentImage::getU(){
	return this->u;
}

/**
 * Crops the latent image to the size given by the input parameter.
 * @param size	The size to which the latent image needs to be cropped.
 */
void LatentImage::cropImage(const cv::Size size){
	this->image = this->image(Rect(0,0,size.width, size.height));
}


/**
 * Calculates the nominator part 1 used in the latent image calculation.
 * It consists of the conjugate of the FFT of the kernel,
 * multiplied by the fft of the blurred image.
 *
 * @param Kfft	a Mat object representing the FFT of the kernel.
 * @param b		a Mat object representing the blurred image.
 *
 * @return a Mat calculated as  conj(Kfft).*fft(b)
 */
Mat LatentImage::calculateNomin1(cv::Mat& Kfft, cv::Mat& b){
  //nomin1 ← conj(Kfft).*fft(b)
  Mat nomin1;
  Mat temp = ImageParameter::fft2(b);
  Mat nominTemp = ImageParameter::conj(Kfft);
  mulSpectrums(nominTemp, temp, nomin1, 0, false );
  return nomin1;
}

/**
 * Calculates the nominator part 2 used in the latent image calculation.
 * It consists of abs(psf2otf(fx,size))²+abs(psf2otf(fy,size))²
 * where fx is a gradient filter in the x direction and fy a gradient filter in the y direction.
 *
 * @param size	size to be used for the psf2otf function.
 * @return	a Mat object calculated as abs(psf2otf(fx,size))²+abs(psf2otf(fy,size))²
 *
 * @throw domain_error if size < 2
 */
Mat LatentImage::calculateDenomin2(cv::Size size) throw (domain_error){
  //denomin2 ← abs(psf2otf(fx,size))²+abs(psf2otf(fy,size))²

  if(size.height <2 || size.width < 2){
	throw domain_error("LatentImage::calculateDenomin2: size needs to greater then 1");
  }

  Mat fx = (Mat_<double>(1,2) << 1, -1);
  Mat fy = (Mat_<double>(2,1) << 1, -1);

  Mat otfFx = psf2otf(fx, size);
  Mat otfFy = psf2otf(fy, size);
  otfFx = ImageParameter::absModules(otfFx);
  otfFy = ImageParameter::absModules(otfFy);
  pow(otfFx, 2, otfFx);
  pow(otfFy, 2, otfFy);
  Mat denomin2 = otfFx + otfFy;

  return denomin2;
}

/**
 * applied a bilateral filter to the input image img.
 * The sigma for the color similary is given by the input parameter sigma.
 * The sigma for the spacial similarity is give nby the input parameter sigmaS.
 * For both filtering in the spacitial and similarity domain, a gaussian distribution is used.
 *
 * @param img		a Mat object on which the bilateral filtering is applied.
 * @param sigmaS 	the value for sigma used for filtering in the spatial domain.
 * @param sigma		the value for sigma used for filtering in the similarity domain.
 *
 * @return the filtered img object.
 * @throw domain_error if img depth is not 64 bits or when sigma or sigmaS <0.
 */
Mat LatentImage::biLateralFilterPan(cv::Mat& img, double sigmaS, double sigma) throw (domain_error){

  if(sigmaS <=0 || sigma <=0 ){
    throw domain_error("LatentImage::bilateralFilterPan: sigmaS and sigma need to be greater then zero");
  }
  if (img.depth() != CV_64F && img.depth() != CV_32F){
	throw domain_error("LatentImage::bilateralFilterPan: bilateral filter Pan needs to be applied to depth CV_64F");
  }

  int channels = img.channels();
  Size size = img.size();

  // Set sigma and convert img to lab
  Mat lab;
  if (channels ==3){
	// if channels =3 then img can be converted to lab colors but cvtColor doesn't work no 64F depth
	if(img.depth() == CV_64F){
		img.convertTo(lab, CV_32F);
		cvtColor(lab, lab, CV_BGR2Lab);
		lab.convertTo(lab, CV_64F);
	} else {
		cvtColor(img, lab, CV_BGR2Lab);
	}
    sigma = sigma *100;
  } else {
	lab = img;
	sigma = sigma * sqrt(channels);
  }


  // set padding
  int fr = ceil(sigmaS*3);
  Mat pImg, pLab;
  copyMakeBorder( img, pImg, fr, fr, fr, fr,  BORDER_REPLICATE);
  copyMakeBorder( lab, pLab, fr, fr, fr, fr,  BORDER_REPLICATE);

  Mat wSum = Mat::zeros(size, CV_64F);
  Mat planes[4];
  for (int ch=0; ch<channels; ch++){
    wSum.copyTo(planes[ch]);
  }
  Mat rImg;
  merge(planes, channels, rImg);
  rImg.copyTo(rImg);

  // crease 2d gaussian kernel
  Mat kernelX = getGaussianKernel((2*fr+1), sigmaS, CV_64F);
  Mat kernelY = getGaussianKernel((2*fr+1), sigmaS, CV_64F);
  Mat spatialWeight = kernelX * kernelY.t();

  double ss = pow(sigma,2);

  Mat fDiff, fDist;

  for (int y=-fr; y<(fr+1); y++){
    for (int x=-fr; x<(fr+1); x++){
      double weight = spatialWeight.at<double>(y+fr,x+fr);

      Mat nImg = pImg(Rect(fr+x, fr+y, size.width, size.height));
      Mat nLab = pLab(Rect(fr+x, fr+y, size.width, size.height));

        lab.copyTo(fDiff);
      fDiff = fDiff - nLab;

      Mat planesFDiff[4];
      split(fDiff, planesFDiff);
      for (int ch=0; ch<channels; ch++){
        if(ch==0){
        	planesFDiff[ch].copyTo(fDist);
    	  pow(fDist,2, fDist);
    	} else {
    	  Mat temp;
		  pow(planesFDiff[ch],2,temp);
    	  fDist = fDist + temp;
    	}
      }

      Mat wf = -.5 * fDist / ss;
      cv::exp(wf,wf);
      wf = wf * weight;


      for (int ch=0; ch<channels; ch++){
    	planes[ch] = wf;
      }
      Mat wf3d;
      merge(planes, channels, wf3d);

      rImg =rImg + nImg.mul(wf3d);

       wSum = wSum + wf;

    }
  }

  // create 3d variant of wSum
  for (int ch=0; ch<channels; ch++){
     wSum.copyTo(planes[ch]);
  }
  Mat wSum3d;
  merge(planes, channels, wSum3d);

  rImg = rImg/(wSum3d);

  return rImg;
}

