/*
 * DeblurAdapter.cc
 *
 * @author  T. Poppe
 * @author  E. van Schaijk
 * @version  1.0
 * @since  July, 2016
 *
 * This module is an adapter module between the gui modules written in c
 * and the domain layer written in CPP. It instantiates the main controller
 * from the domain layer and passed the input parameters from the module
 * DeblurPanPlugin. The results are passed to the module ResultsDialog for user
 * validation before it is passed back to DeblurPanPlugin.
 *
 */

#include "DeblurAdapter.hpp"
extern "C" {
  #include <libgimp/gimp.h>
  #include <libgimp/gimpui.h>
  #include <log4c.h>
  #include <glib.h>
  #include <gdk-pixbuf/gdk-pixbuf.h>
  #include "ResultsDialog.h"
}
#include <opencv2/core/core.hpp>
#include <opencv2/opencv.hpp>
#include <stdexcept>
#include "../textDeblurPan/FrameworkController.hpp"
#include "ProgressBarUpdater.hpp"

using namespace cv;
using namespace std;

// Declaring the local procedures
static cv::Mat checkMat(Mat src, int height, int weight, int channels);
static GdkPixbuf* toPixbuf(cv::Mat &mat);
static DeblurParameters copyDeblurParamStruct(DeblurParam deblurParam);
static void showErrorMessage(domain_error& e);

/**
 * Deblurs the blurred image with the text deblur algoritm of Pan et all.
 * with the deblur parameters settings as stored in the struct deblurParam
 * The blurred image is part of the deblurImageData . The result is stored
 * back in the deblurImageData.latent.
 *
 * @param deblurParam	A DeblurParam struct containing the deblur parameters which are used for deblurring
 * @param deblurImageData	A struct holding the pointers for the Gim pixel region for the blurred image
 * 							and for the result of the deblur
 *
 * 	@return		zero if the user cancels the deblur in the results dialog, -1 in case of an error otherwise one
 */
int deblurFromGimp(DeblurParam deblurParam, DeblurImageData deblurImageData){

  log4c_category_t* mycat2 = log4c_category_get("deblur.pan.overgang");
  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"starting conversion to mat objects");
  gint res = 1;// return value

  // get image data
  gint x = deblurImageData.x;
  gint y = deblurImageData.y;
  gint height = deblurImageData.imageHeight;
  gint width = deblurImageData.imageWidth;
  gint channels = gimp_drawable_bpp((deblurImageData.drawable)->drawable_id);

  // copy pixelregion to mat
  Mat temp = Mat( height, width, CV_MAKETYPE(CV_8U, channels));
  gimp_pixel_rgn_get_rect (deblurImageData.blurredPixelRgn, temp.data ,x,y, width,height);
  cv::Mat blurredImage = temp.clone(); //clones the temp into a new mat blurred Image to avoid interference with Gimp

  // copy pixelregion to pixbuf for the resultDialog
  deblurImageData.blurredPixbuf = toPixbuf(blurredImage);
  Mat latentMat, IntermediaryMat, kernelMat;

  // copy the deblur parameter settings to the deblurController required format
  DeblurParameters deblurParameters = copyDeblurParamStruct(deblurParam);

  // Retrieve an instance of the framework controller
  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"instantiate framworkcontroller");
  FrameworkController& fwc = FrameworkController::getInstance();

  // create progressBarUpdater object and register it as observer at the frameworkcontroller
  ProgressBarUpdater progressBarUpdater = NULL;
  if (deblurImageData.runMode == preview){
	  progressBarUpdater = ProgressBarUpdater(deblurParam.progressPreview);
	  fwc.attach(&progressBarUpdater);
  }else{
	  progressBarUpdater = ProgressBarUpdater(deblurParam.progressMain);
	  fwc.attach(&progressBarUpdater);
  }

  // deblur image
  try {
    fwc.deblur(deblurParameters, blurredImage, kernelMat, IntermediaryMat, latentMat);
  } catch(domain_error& e){
	  showErrorMessage(e);
	  return -1;
  }
  // remove ProgressBarUpdater from the observer list
  fwc.detach(&progressBarUpdater);

  // Validate latent image and set rect in pixelbuf
  Mat latentMatChecked = checkMat(latentMat, height, width, channels);
  gimp_pixel_rgn_set_rect (deblurImageData.latentPixelRgn, latentMatChecked.data, x,y,width,height);
  deblurImageData.latentPixbuf = toPixbuf(latentMatChecked);

  // set kernel rect in pixel region
  cv::cvtColor(  kernelMat,kernelMat,CV_GRAY2RGB);
  Mat kernelMatChecked = checkMat(kernelMat, kernelMat.rows, kernelMat.cols,3);
  deblurImageData.kernelPixbuf = toPixbuf(kernelMatChecked);

  // set intermediary image in pixelbuf
  cv::cvtColor(  IntermediaryMat,IntermediaryMat,CV_GRAY2RGB);
  deblurImageData.intermediaryPixbuf = toPixbuf(IntermediaryMat);

  // show kernel and intermediate results in case of interactive runmode
  if(deblurImageData.runMode == interactive){
	res = showResults(deblurImageData);
  }

  // set progresbar back to zero
  if(deblurImageData.runMode != nonInteractive){
    progressBarUpdater.update(0);
  }

  return res;

}

/**
 * Retrieve the maximum kernelsize given the x and y dimensions of the blurred image.
 * The kernel is assumed to be a square, i.e. width is equal to the height.
 *
 * @param x	the width of the blurred image
 * @param y the height of the blurred image
 * @return	The maximum size of the kernel
 */
int getMaxKernelSize(gint x, gint y){
	FrameworkController& fwc = FrameworkController::getInstance();
	int size = fwc.getMaxKernelSize(Size(x,y));
	return size;
};


/**
 * Converts an picture from OpenCV mat format into GdkPixBuf and resized the picture format to
 * max 200 pix keepting cols/width ratio unchanged. Opencv mat picture needs to have depth 8U
 * and can have one of the following channels combinations:
 * 		rgb, rgba, grayscale and grayscale + alpha
 * 	Note that the complete content is copied and not referred
 *
 * 	@param mat 	The picture in OpenCv Mat format with depth u8
 * 	@return 	The picture in GdkPixbuf format resized to max 200 pix keepting cols/width ratio unchanged
 */
static GdkPixbuf* toPixbuf(cv::Mat &mat){
  Mat matResized;
  GdkPixbuf *pixbuf;

  // resize image to max 200 pixel width or heigth
  gint maxSize = std::max(mat.cols, mat.rows);
  cout << " maxSize " << maxSize << endl;
  gdouble factor = 200.0/maxSize;
  cout << " factor is " << factor << endl;
  Size newSize = Size(mat.cols * factor, mat.rows * factor);
  cout << " cols " << newSize.width << " rows " << newSize.height << endl;
  resize(mat,matResized,newSize,0,0,CV_INTER_LINEAR);

  // check if picture is RGB with alhpa channel which needs to be removed
  if (matResized.channels() == 4){
    cvtColor(matResized,matResized, CV_RGBA2RGB);
  }

  // check if picture is grayscale with alhpa channel which needs to be removed
  if(matResized.channels() == 2){
    Mat temp[2];
    split(matResized, temp);
    matResized = temp[0].clone();
  }

  // if picture is grayscale then it needs to be converted to 3 channel BGR
  if (matResized.channels() == 1){
	  Mat temp[3];
	  for (int i =0; i<3;i++){
	    temp[i] = matResized.clone();
	  }
	  merge(temp, 3, matResized);
  }

  pixbuf = gdk_pixbuf_new_from_data(matResized.data,GDK_COLORSPACE_RGB, false, 8, matResized.cols,
			matResized.rows, matResized.step, NULL, NULL);

  return gdk_pixbuf_copy(pixbuf);
}


/**
 * Validates and converts the input Mat on dimensions, number of channels and depth.
 * This procedures is called before the results are set back to Gimp as additional
 * validation step. The depth should always be 8 bits. If theDimensions deviate from
 * the requirements, then the output Mat is converted to meet the requirements.
 * As last step the procedure check whether the source Mat is stored in one continous
 * block, this is requirement as only the pointer to the data block is transferred to
 * Gimp and Gimp assumed continuous storage in one block of the data
 *
 * @param src		The mat which is validated
 * @param height	The required height of the input Mat
 * @param width		The required width of the input Mat
 * @param channels 	The required number of channels of the input Mat
 *
 * @return			The src Mat converted to the height,width, number of channels
 * 						and depth 8 bits as stated in the input parameters.
 */
static cv::Mat checkMat(Mat src, int height, int width, int channels){
  log4c_category_t* mycat2 = log4c_category_get("deblur.pan");
  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"start check methode");

  // use clone here to ensure continous storage of data
  Mat res = src.clone();

  // check  type
  if ( src.type() !=  CV_MAKETYPE(CV_8U, src.channels())){
	  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"type ongelijk");
	  src.convertTo(res,CV_MAKETYPE(CV_8U, src.channels()) );
  }

  // check channels
  if(src.channels() != channels ){
	log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"channels ongelijk");
	// if aplha channels needs to be added
	if(src.channels() ==3 && channels ==4){
	  // src mat is missing an alpha channel
	  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"adding alpha channel");
	  Mat newSrc = Mat(src.rows,src.cols,CV_8UC4);

	  Mat planes[3];
	  split(res, planes);
	  Mat alpha = Mat::ones(src.rows, src.cols, CV_8U)*255;
	  vector<Mat> ch;
	  ch.push_back(planes[0]);
	  ch.push_back(planes[1]);
	  ch.push_back(planes[2]);
	  ch.push_back(alpha);
	  merge(ch, newSrc);
	  newSrc.copyTo(res);
	} else if (src.channels() ==1 && channels >=3){
	  // convert src from gray to color
	  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"convert to color from gray");
	  Mat newSrc;
	  cv::cvtColor(res,newSrc, CV_GRAY2BGR, channels);
	  newSrc.copyTo(res);
	} else  {
	  throw domain_error("number of channels is incorrect");
	}
  }

  // check width and height
  if(src.rows != height || src.cols != width){
	  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"afmeting ongelijk");
	  Mat resized;
	  Size dsize = Size(width, height);
	  cv::resize(res,resized,dsize,0,0,INTER_NEAREST);
	  resized.copyTo(res);
  }

  if(!src.isContinuous()){
	  log4c_category_log(mycat2, LOG4C_PRIORITY_NOTICE,"niet continue");
	  throw domain_error("data blok niet continues");
  }

  return res;
}


/**
 * Copies the parameter values for debluring as set in struct deblurParam via the GUI
 * into the struct deblurParameters which is used by the domain layer.
 * The boolean gamma correction is translated into 1 in case of false and
 * into 2.2 in case of true
 *
 * @param deblurParam	The struct with the parameter values as set in the GUI
 *
 * @return	A struct DeblurParameters with the values copied from the input which is used
 * 				by the domain layer.
 */
static DeblurParameters copyDeblurParamStruct(DeblurParam deblurParam){

  DeblurParameters deblurParameters;

  deblurParameters.kernelSize = deblurParam.kernelSize;
  deblurParameters.numberIterations = deblurParam.numberIterations;
  if (deblurParam.gammaCorrectionBoolean){
    deblurParameters.gammaCorrection = 2.2;
  } else {
    deblurParameters.gammaCorrection = 1;
  }

  deblurParameters.lambdaGrad = deblurParam.lambdaGrad;
  deblurParameters.lambdaPixel = deblurParam.lambdaPixel;
  deblurParameters.lambdaTV = deblurParam.lambdaTV;
  deblurParameters.lambdaL0 = deblurParam.lambdaL0;
  deblurParameters.weightRing = deblurParam.weightRing;

  return deblurParameters;
}


/**
 * Show the error thrown by the domain layer.
 *
 * @param e	The error object from the domain layer.
 */
static void showErrorMessage(domain_error& e){
	GtkDialogFlags flags = GTK_DIALOG_DESTROY_WITH_PARENT;
	GtkWidget* dialog = gtk_message_dialog_new (NULL,
	                                 flags,
	                                 GTK_MESSAGE_ERROR,
	                                 GTK_BUTTONS_CLOSE,
	                                 e.what()
	                                 );
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}
