#include "DeblurDialog.h"
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <glib.h>

#include "DeblurPanPlugin.h"

/*
 * @author  T. Poppe
 * @author  E. van Schaijk
 * @version  1.0
 * @since  July, 2016
 *
 * Module for displaying the GUI which is used for setting the deblur parameter values.
 */

/* declaration of Local function prototypes  */
static void checkEvent (GtkWidget *widget, gint *data); // function for handling checkbox toggle event

/**
 * Shows the dialog in which the user can enter the deblur parameters
 * The values are stored in the struct DeblurParam. The dialog return a response id
 * equal to the GTK response id version 2.
 *
 *
 * @param deblur_param	A pointer to a DeblurParam struct which is used to store
 * 				the parameter values set in the GUI
 * @param deblurParamDefaults	A pointer  to a DeblurParam struct containing the default
 *	 	 	 	 	 	 	 	 setting for deblurring.
 * @param drawable 		A pointer to the GimpDrawable containing the deblurred picture from GIMP
 *
 * @return	A boolean indicating the status.
 * 				-5 = status OK
 * 				-6 = status cancel
 *
 */
gboolean
blur_dialog(
  DeblurParam *deblur_param,
  DeblurParam *deblurParamDefaults,
  GimpDrawable *drawable)
{
  GtkWidget *dialog;
  GtkWidget *preview;
  GtkWidget *preview_update_button;
  GtkWidget *kernel_frame, *decon_frame;
  GtkWidget *main_vbox, *kernelTable, *deconTable;
  GtkWidget *left_vbox, *right_vbox, *main_hbox, *button_hbox;
  GtkWidget *rightAlignment, *leftAlignment, *progressAlignment;
  GtkWidget *progressMain, *progressPreview;

  //slider controls
  GtkWidget *kernelSlider, *lambdaGradSlider, *lambdaPixelSlider, *numberIterSlider, *dummyLabel,
  	  	  *lambdaTVSlider, *lambdaL0Slider, *dummyLabel2;
  // defone labels for the parameters
  GtkWidget *kernel_label, *lambda_pixel_label, *lambda_grad_label, *number_iter_label,
		*lambda_tv_label, *lambda_l0_label;
  // define controls for the parameters
  GtkWidget *kernel_spinbutton, *gamma_checkbox, *lambda_pixel_spinbutton,
		*lambda_grad_spinbutton, *number_iter_spinbutton, *lambda_tv_spinbutton,
		*lambda_l0_spinbutton, *weight_ring_checkbox;
  // define handles for values changes of the parameters
  GtkObject *kernel_spinbutton_adj, *lambda_pixel_spinbutton_adj,
		*lambda_grad_spinbutton_adj, *number_iter_spinbutton_adj, *lambda_tv_spinbutton_adj,
		*lambda_l0_spinbutton_adj;
  gboolean   run=0, response = 0;


  gimp_ui_init ("deblurPan", FALSE);

  dialog = gimp_dialog_new ("Deblur Pan et al", "deblur",
		  NULL, 0,gimp_standard_help_func, "plug-in-template", "Reset", 1, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
          GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);

  // create the main box holding all gui components.
  // deze is gemaakt om evt later de voortgangsbalk toe te kunnen voegen
  main_vbox = gtk_vbox_new (FALSE, 6);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), main_vbox);
  gtk_widget_show (main_vbox);

  // create main hbox for left and right screen
  main_hbox = gtk_hbox_new(FALSE,6);
  gtk_container_add (GTK_BOX (main_vbox), main_hbox);
  gtk_widget_show (main_hbox);

  // add main progress bar
  progressAlignment = gtk_alignment_new (0, 0, 1, 1);
  gtk_widget_show (progressAlignment);
  gtk_box_pack_start (GTK_BOX (main_vbox), progressAlignment, FALSE, FALSE, 0);
  gtk_alignment_set_padding (GTK_ALIGNMENT (progressAlignment), 10, 10, 10, 10);

  progressMain = gtk_progress_bar_new ();
  gtk_widget_show (progressMain);
  gtk_container_add (GTK_CONTAINER (progressAlignment), progressMain);

  // create left box alignment
  leftAlignment = gtk_alignment_new (0, 0, 1, 1);
  gtk_widget_show (leftAlignment);
  gtk_box_pack_start (GTK_BOX (main_hbox), leftAlignment, FALSE, FALSE, 0);
  gtk_alignment_set_padding (GTK_ALIGNMENT (leftAlignment), 10, 10, 10, 10);

  //create left vbox
  left_vbox = gtk_vbox_new(FALSE,6);
  gtk_container_add (GTK_CONTAINER (leftAlignment), left_vbox);
  gtk_widget_show (left_vbox);

  // create right_vbox_alignment
  rightAlignment = gtk_alignment_new (0, 0, 1, 1);
  gtk_widget_show (rightAlignment);
  gtk_box_pack_start (GTK_BOX (main_hbox), rightAlignment, FALSE, FALSE, 0);
  gtk_alignment_set_padding (GTK_ALIGNMENT (rightAlignment), 10, 10, 10, 10);

  //create right vbox
  right_vbox = gtk_vbox_new(FALSE,6);
  gtk_container_add (GTK_CONTAINER (rightAlignment), right_vbox);
  gtk_widget_show (right_vbox);

  // create preview window
  preview = gimp_drawable_preview_new (drawable, &(deblur_param->preview));
  gimp_preview_set_update((struct _GimpPreview *)preview, FALSE);
  gtk_box_pack_start (GTK_BOX (left_vbox), preview, FALSE, FALSE, 0);
  gtk_widget_show (preview);

  // add progress bar for preview
  progressPreview = gtk_progress_bar_new ();
  gtk_widget_show (progressPreview);
  gtk_box_pack_start (GTK_BOX (left_vbox), progressPreview, FALSE, FALSE, 0);

  // create box for button
  button_hbox = gtk_hbox_new (FALSE, 1);
  gtk_misc_set_alignment(GTK_MISC(button_hbox),0,0);
  gtk_container_set_border_width (GTK_CONTAINER (button_hbox), 1);
  gtk_widget_show (button_hbox);
  gtk_box_pack_start (GTK_BOX (left_vbox), button_hbox, FALSE, FALSE, 0);

  preview_update_button = gtk_button_new_with_mnemonic ("update preview");
  gtk_box_pack_start (GTK_BOX (button_hbox), preview_update_button, FALSE, FALSE, 0);
  gtk_widget_show (preview_update_button);


  // create kernel frame to hold kernel parameter controls
  kernel_frame = gimp_frame_new ("Kernel estimation");
  gtk_box_pack_start (GTK_BOX (right_vbox), kernel_frame, FALSE, FALSE, 10);
  gtk_widget_show (kernel_frame);

  kernelTable = gtk_table_new(5,3,FALSE);
  gtk_widget_show (kernelTable);
  gtk_container_add (GTK_CONTAINER (kernel_frame), kernelTable);
  gtk_table_set_col_spacing ( GTK_TABLE(kernelTable),1,15);
  gtk_table_set_col_spacing ( GTK_TABLE(kernelTable),0,15);

  // create deconvolution frame to hold final deconvolution parameter controls
  decon_frame = gimp_frame_new ("Final Deconvolution");
  gtk_box_pack_start (GTK_BOX (right_vbox), decon_frame, FALSE, FALSE, 0);
  gtk_widget_show (decon_frame);

  deconTable = gtk_table_new(3,3,FALSE);
  gtk_widget_show (deconTable);
  gtk_container_add (GTK_CONTAINER (decon_frame), deconTable);
  gtk_table_set_col_spacing ( GTK_TABLE(deconTable),1,15);
  gtk_table_set_col_spacing ( GTK_TABLE(deconTable),0,15);



  // add controls for kernel size
  kernel_label = gtk_label_new_with_mnemonic ("Kernel size");
  gtk_widget_show (kernel_label);
  gtk_misc_set_alignment(GTK_MISC(kernel_label),0.0,0.5);
  gtk_table_attach(GTK_TABLE(kernelTable), kernel_label,0,1,0,1, GTK_FILL, GTK_FILL, 0, 5);


  kernel_spinbutton = gimp_spin_button_new (&kernel_spinbutton_adj, deblur_param->kernelSize,
		  KERNELMINSIZE, deblur_param->kernelMaxSize, 2, 10, 1, 5, 0);
  gtk_misc_set_alignment(GTK_MISC(kernel_spinbutton),1,1);
  gtk_table_attach(GTK_TABLE(kernelTable), kernel_spinbutton, 2,3,0,1,GTK_FILL,GTK_FILL,0,5);
  gtk_widget_show (kernel_spinbutton);

  kernelSlider = gtk_hscale_new(GTK_ADJUSTMENT(kernel_spinbutton_adj));
  gtk_widget_show (kernelSlider);
  gtk_scale_set_draw_value(GTK_SCALE(kernelSlider), 0 );
  gtk_table_attach(GTK_TABLE(kernelTable), kernelSlider,1,2,0,1, GTK_FILL, GTK_FILL, 0, 5);

  // dummy label to control width of slider
  dummyLabel = gtk_label_new_with_mnemonic("                                                                        ");
  gtk_widget_show (dummyLabel);
  gtk_table_attach(GTK_TABLE(kernelTable), dummyLabel,1,2,4,5, GTK_FILL, GTK_FILL, 0, 5);

  // add controls for lambda pixel
  lambda_pixel_label = gtk_label_new_with_mnemonic ("Lambda pixel");
  gtk_widget_show (lambda_pixel_label);
  gtk_misc_set_alignment(GTK_MISC(lambda_pixel_label),0.0,0.5);
  gtk_table_attach(GTK_TABLE(kernelTable), lambda_pixel_label, 0,1,1,2,GTK_FILL, GTK_FILL, 0,5);

  lambda_pixel_spinbutton = gimp_spin_button_new (&lambda_pixel_spinbutton_adj, deblur_param->lambdaPixel,
		  LAMBDAPIXELMIN, LAMBDAPIXELMAX, 0.00001, 1, 1, 5, 5);
  gtk_table_attach(GTK_TABLE(kernelTable), lambda_pixel_spinbutton, 2,3,1,2,GTK_FILL, GTK_FILL,0,5);
  gtk_widget_show (lambda_pixel_spinbutton);

  lambdaGradSlider = gtk_hscale_new(GTK_ADJUSTMENT(lambda_pixel_spinbutton_adj));
  gtk_widget_show (lambdaGradSlider);
  gtk_scale_set_draw_value(GTK_SCALE(lambdaGradSlider), 0 );
  gtk_table_attach(GTK_TABLE(kernelTable), lambdaGradSlider,1,2,1,2, GTK_FILL, GTK_FILL, 0, 5);

  // add controls for lambda grad
  lambda_grad_label = gtk_label_new_with_mnemonic ("lambda grad :");
  gtk_widget_show (lambda_grad_label);
  gtk_misc_set_alignment(GTK_MISC(lambda_grad_label),0.0,0.5);
  gtk_table_attach(GTK_TABLE(kernelTable), lambda_grad_label, 0,1,2,3,GTK_FILL, GTK_FILL,0,5);
  gtk_label_set_justify (GTK_LABEL (lambda_grad_label), GTK_JUSTIFY_RIGHT);

  lambda_grad_spinbutton = gimp_spin_button_new (&lambda_grad_spinbutton_adj, deblur_param->lambdaGrad,
		  LAMBDAGRADMIN, LAMBDAGRADMAX, 0.00001,1, 1, 5, 5);
  gtk_table_attach(GTK_TABLE(kernelTable), lambda_grad_spinbutton, 2,3,2,3,GTK_FILL, GTK_FILL,0,5);
  gtk_widget_show (lambda_grad_spinbutton);

  lambdaPixelSlider = gtk_hscale_new(GTK_ADJUSTMENT(lambda_grad_spinbutton_adj));
  gtk_widget_show (lambdaPixelSlider);
  gtk_scale_set_draw_value(GTK_SCALE(lambdaPixelSlider), 0 );
  gtk_table_attach(GTK_TABLE(kernelTable), lambdaPixelSlider,1,2,2,3, GTK_FILL, GTK_FILL, 0, 5);

  // add controls for number of iterations
  number_iter_label = gtk_label_new_with_mnemonic ("number of iterations :");
  gtk_widget_show (number_iter_label);
  gtk_misc_set_alignment(GTK_MISC(number_iter_label),0.0,0.5);
  gtk_table_attach(GTK_TABLE(kernelTable), number_iter_label, 0,1,3,4,GTK_FILL, GTK_FILL,0,5);
  gtk_label_set_justify (GTK_LABEL (number_iter_label), GTK_JUSTIFY_RIGHT);

  number_iter_spinbutton = gimp_spin_button_new (&number_iter_spinbutton_adj, deblur_param->numberIterations,
		  NUMBERITERMIN, NUMBERITERMAX, 1, 1, 1, 5, 0);
  gtk_table_attach(GTK_TABLE(kernelTable), number_iter_spinbutton, 2,3,3,4,GTK_FILL, GTK_FILL,0,5);
  gtk_widget_show (number_iter_spinbutton);

  numberIterSlider = gtk_hscale_new(GTK_ADJUSTMENT(number_iter_spinbutton_adj));
  gtk_widget_show (numberIterSlider);
  gtk_scale_set_draw_value(GTK_SCALE(numberIterSlider), 0 );
  gtk_table_attach(GTK_TABLE(kernelTable), numberIterSlider,1,2,3,4, GTK_FILL, GTK_FILL, 0, 5);

  // add controls for gamma correction
  gamma_checkbox = gtk_check_button_new_with_mnemonic ("_Gamma correction");
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (gamma_checkbox), deblur_param->gammaCorrectionBoolean);
  gtk_widget_show (gamma_checkbox);
  gtk_misc_set_alignment(GTK_MISC(gamma_checkbox),0.0,0.5);
  gtk_table_attach(GTK_TABLE(kernelTable), gamma_checkbox,0,1,4,5, GTK_FILL, GTK_FILL, 0, 5);

  // add controls for lambda TV
  lambda_tv_label = gtk_label_new_with_mnemonic ("lambda TV :");
  gtk_widget_show (lambda_tv_label);
  gtk_misc_set_alignment(GTK_MISC(lambda_tv_label),0.0,0.5);
  gtk_table_attach(GTK_TABLE(deconTable), lambda_tv_label,0,1,0,1, GTK_FILL, GTK_FILL, 0, 5);
  gtk_label_set_justify (GTK_LABEL (lambda_tv_label), GTK_JUSTIFY_RIGHT);

  lambda_tv_spinbutton = gimp_spin_button_new (&lambda_tv_spinbutton_adj, deblur_param->lambdaTV,
		  LAMBDATVMIN, LAMBDATVMAX, 0.00001,1, 1, 5, 5);
  gtk_table_attach(GTK_TABLE(deconTable), lambda_tv_spinbutton, 2,3,0,1,GTK_FILL, GTK_FILL,0,5);
  gtk_widget_show (lambda_tv_spinbutton);

  lambdaTVSlider = gtk_hscale_new(GTK_ADJUSTMENT(lambda_tv_spinbutton_adj));
  gtk_widget_show (lambdaTVSlider);
  gtk_scale_set_draw_value(GTK_SCALE(lambdaTVSlider), 0 );
  gtk_table_attach(GTK_TABLE(deconTable), lambdaTVSlider,1,2,0,1, GTK_FILL, GTK_FILL, 0, 5);

  // add controls for lambda L0
  lambda_l0_label = gtk_label_new_with_mnemonic ("lambda l0 :");
  gtk_widget_show (lambda_l0_label);
  gtk_misc_set_alignment(GTK_MISC(lambda_l0_label),0.0,0.5);
  gtk_table_attach(GTK_TABLE(deconTable), lambda_l0_label,0,1,1,2, GTK_FILL, GTK_FILL, 0, 5);
  gtk_label_set_justify (GTK_LABEL (lambda_l0_label), GTK_JUSTIFY_RIGHT);


  lambda_l0_spinbutton = gimp_spin_button_new (&lambda_l0_spinbutton_adj, deblur_param->lambdaL0,
		  LAMBDAGRADMIN, LAMBDAGRADMAX, 0.00001,1, 1, 5, 5);
  gtk_table_attach(GTK_TABLE(deconTable), lambda_l0_spinbutton, 2,3,1,2,GTK_FILL, GTK_FILL,0,5);
  gtk_widget_show (lambda_l0_spinbutton);

  lambdaL0Slider = gtk_hscale_new(GTK_ADJUSTMENT(lambda_l0_spinbutton_adj));
  gtk_widget_show (lambdaL0Slider);
  gtk_scale_set_draw_value(GTK_SCALE(lambdaL0Slider), 0 );
  gtk_table_attach(GTK_TABLE(deconTable), lambdaL0Slider,1,2,1,2, GTK_FILL, GTK_FILL, 0, 5);

  // add controls for weight ringing
  weight_ring_checkbox = gtk_check_button_new_with_mnemonic ("_Ringing removal");
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (weight_ring_checkbox), deblur_param->weightRing);
  gtk_widget_show (weight_ring_checkbox);
  gtk_misc_set_alignment(GTK_MISC(weight_ring_checkbox),0.0,0.5);
  gtk_table_attach(GTK_TABLE(deconTable), weight_ring_checkbox,0,1,2,3, GTK_FILL, GTK_FILL, 0, 5);

  // dummy label to determine width of slider
  dummyLabel2 = gtk_label_new_with_mnemonic("                                                                        ");
  gtk_widget_show (dummyLabel2);
  gtk_table_attach(GTK_TABLE(deconTable), dummyLabel2,1,2,2,3, GTK_FILL, GTK_FILL, 0, 5);


  // connect preview to button
  g_signal_connect_swapped (preview, "invalidated",
                            G_CALLBACK (deblurPreview),
							deblur_param);
  g_signal_connect_swapped(G_OBJECT(preview_update_button), "clicked",
                             G_CALLBACK (gimp_preview_invalidate),
                             preview);

  // connect spinbuttons with deblur_param variables
  g_signal_connect (kernel_spinbutton_adj, "value_changed",
                      G_CALLBACK (gimp_int_adjustment_update),
					  &(deblur_param->kernelSize));
  g_signal_connect (lambda_pixel_spinbutton_adj, "value_changed",
                      G_CALLBACK (gimp_double_adjustment_update),
					  &(deblur_param->lambdaPixel));
  g_signal_connect (lambda_grad_spinbutton_adj, "value_changed",
                      G_CALLBACK (gimp_double_adjustment_update),
					  &(deblur_param->lambdaGrad));
  g_signal_connect (number_iter_spinbutton_adj, "value_changed",
                      G_CALLBACK (gimp_int_adjustment_update),
					  &(deblur_param->numberIterations));
  g_signal_connect (lambda_tv_spinbutton_adj, "value_changed",
                      G_CALLBACK (gimp_double_adjustment_update),
					  &(deblur_param->lambdaTV));
  g_signal_connect (lambda_l0_spinbutton_adj, "value_changed",
                      G_CALLBACK (gimp_double_adjustment_update),
					  &(deblur_param->lambdaL0));

  // connect checkboxes
  gtk_signal_connect (GTK_OBJECT (weight_ring_checkbox), "toggled",
 			GTK_SIGNAL_FUNC (checkEvent), &(deblur_param->weightRing));
  gtk_signal_connect (GTK_OBJECT (gamma_checkbox), "toggled",
 			GTK_SIGNAL_FUNC (checkEvent), &(deblur_param->gammaCorrectionBoolean));

  // set progressbar in deblurParam for updating
  deblur_param->progressMain = GTK_PROGRESS_BAR(progressMain);
  deblur_param->progressPreview = GTK_PROGRESS_BAR(progressPreview);

  gtk_widget_show (dialog);

  //run = (gimp_dialog_run (GIMP_DIALOG (dialog)) == GTK_RESPONSE_OK);

  while (response >= 0 ){
	  response = gimp_dialog_run(GIMP_DIALOG (dialog));
	  if (response == 1){
		  //reset the value to default
		  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (weight_ring_checkbox), deblurParamDefaults->weightRing);
		  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON (gamma_checkbox), deblurParamDefaults->gammaCorrectionBoolean);
		  gtk_adjustment_set_value (GTK_ADJUSTMENT(kernel_spinbutton_adj), deblurParamDefaults->kernelSize);
		  gtk_adjustment_set_value (GTK_ADJUSTMENT(lambda_l0_spinbutton_adj), deblurParamDefaults->lambdaL0);
		  gtk_adjustment_set_value (GTK_ADJUSTMENT(lambda_pixel_spinbutton_adj), deblurParamDefaults->lambdaPixel);
		  gtk_adjustment_set_value (GTK_ADJUSTMENT(lambda_grad_spinbutton_adj), deblurParamDefaults->lambdaGrad);
		  gtk_adjustment_set_value (GTK_ADJUSTMENT(number_iter_spinbutton_adj), deblurParamDefaults->numberIterations);
		  gtk_adjustment_set_value (GTK_ADJUSTMENT(lambda_tv_spinbutton_adj), deblurParamDefaults->lambdaTV);
		  gimp_preview_set_update((struct _GimpPreview *)preview, FALSE);
	  }
	  run = (response == GTK_RESPONSE_OK);
  }
  //gtk_widget_destroy (dialog);

  return run;

}


/**
 * procedure to store the checkEvent of the GTKToggleButtons
 * The event is stored as integer an the adres given by the data pointer.
 *
 * @param widget A pointer to the widget firing the event
 * @param data	 A pointer to the integer wher the event is stored.
 *
 */
void checkEvent (GtkWidget *widget, gint *data)
{
    *data = gtk_toggle_button_get_active((struct _GtkToggleButton *)widget);
}
