 /*
 * DeblurPanPlugin.c
 *
 * @author  T. Poppe
 * @author  E. van Schaijk
 * @version  1.0
 * @since  July, 2016
 *
 * This module is the main GUI module. It is invoked by GIMP when calling the
 * plug-in in GIMP. It manages the communication and registration with GIMP,
 * as well as the dialog modules.
 */

#include "DeblurPanPlugin.h"

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <glib.h>
#include <log4c.h>

#include "DeblurAdapter.hpp"
#include "DeblurDialog.h"
#include "ResultsDialog.h"


/* declaration of Local function prototypes  */
static void   query (void);
static void   run   (const gchar *name, gint nparams, const GimpParam *param,
		     	 	 gint *nreturn_vals, GimpParam **return_vals);
static void  initLogging (void);
static void  closeLogging (void);
static void logDeblurParam(DeblurParam deblurParam);
static void initializePixelregions(gint x1,	gint y1, gint width, gint height);
static void mapInputParam(DeblurParam *deblurParam, GimpParam *param);

/* Set up default values for deblur parameters */
DeblurParam  deblurParamDefaults = {
  25,  /* kernelsize */
  0.004, /*lambda_pixel */
  0.004, /*lambda_grad */
  5, /* # iterations */
  FALSE,	 /* preview */
  0.003, /*lambdaTV */
  0.001, /*lambda L0 */
  1, /*weight Ringing */
  NULL, NULL, //progressbars
  FALSE, //gammaCorrectionBoolean
  100 //kernelMaxSize
};

static DeblurParam  deblurParam; // global variabele to hold the deblur parameters
static log4c_category_t* mycat = NULL; // global variable to hold the logging category

/* create deblur result parameter */
static DeblurImageData deblurImageData = {
  0, // image_height
  0, //   image_width
  0, // x
  0, // y
  NULL, // drawable
  FALSE, // preview
  NULL, // latentPixelRgn
  NULL, // kernelPixelRgn
  NULL, // intermediaryPixelRgn
  NULL, // blurredPixelRgn
  interactive
};


/**
 * global variable for the latent and blurred pixel regions.
 * These need to be global to avoid that their data storage is released
 * in the memory
 */
static GimpPixelRgn latentPixelRgn;
static GimpPixelRgn blurredPixelRgn;

/* Implementation of plug-in stub as by convention of GIMP*/
GimpPlugInInfo PLUG_IN_INFO = {
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};



/* required for plug-in */
MAIN ()


/*
 * procedure for procedure registration and input arguments definition
 * Refer to GIMP documentation for details
 */
static void query (void){
  gchar *help_path;
  gchar *help_uri;

  help_path = g_build_filename (DATADIR, "help", NULL); // DATADIR is set during ./config
  help_uri = g_filename_to_uri (help_path, NULL, NULL);
  g_free (help_path);
  // defining the parameters of the plugin which are used in the non interactive mode
  static GimpParamDef args[] = {
    {GIMP_PDB_INT32, "run-mode",  "Run mode"}, /*args[0] will return the run mode */
    {GIMP_PDB_IMAGE, "image", "Input image"}, /*args[1] will return the original image */
    {GIMP_PDB_DRAWABLE, "drawable", "Input drawable"}, /*args[2] will return the drawable */
	{GIMP_PDB_INT32, "kernelSize", "Kernel size"},
	{GIMP_PDB_INT32, "iterations", "Iterations"},
	{GIMP_PDB_FLOAT, "lambdaGrad", "Lambda Gradient"},
	{GIMP_PDB_FLOAT, "lambdaPixel", "Lambda Pixel"},
	{GIMP_PDB_FLOAT, "lambdaTV", "Lambda TV"},
	{GIMP_PDB_FLOAT, "lambdaL0", "Lambda L0"},
	{GIMP_PDB_INT8, "ringing", "Ringing"},
	{GIMP_PDB_INT8, "gammaCorrection", "Gamma Corrections"}
  };

  gimp_plugin_help_register ("http://developer.gimp.org/plug-in-template/help",
                             help_uri);

  // register the plugin in the PCB
  gimp_install_procedure (
    "TextDeblur Pan et al", /* Name */
    "Deblurs text with the Pan et al algorithm", /*Desciption */
    "To do help text",
    "Ton Poppe / Else van Schaijk",
    "Copyright GNU",
    "2015",
    "Text Deblur Pan", /* Menu label */
    "RGB*, GRAY*",
    GIMP_PLUGIN,
    G_N_ELEMENTS (args), 0,
    args, NULL
  );

  // register the plugin in the menu of Gimp
  gimp_plugin_menu_register ("TextDeblur Pan et al", "<Image>/Filters/");
}

/*
 * Main procedure called by GIMP when the function is used in GIMP.
 * The procedure deblurs the drawing which is part of the input array
 * according to the algoritmen of Pan et al.as described in the paper:
 * Jinshan Pan, Zhe Hu, ZhixiJinshan Pan, Zhe Hu, Zhixun Su, and Ming-Hsuan Yang,
 * "Deblurring Text Images via L0-Regularized Intensity and Gradient Prior",
 * CVPR 2014.
 *
 * The settings used for deblurring can be entered in a GUI window.
 * refer to the GIMP plug-in documentation for details on the
 * parameters.
 *
 * The deblur results are sent back to GIMP where it will replace the blurred drawing in GIMP.
 *
 * @param name	the procedure's name.
 * @param nparams	number of input params
 * @param param		input param array
 * @param nreturn_vals	pointer for setting the number of output params
 * @param return_vals	pointer to output param array
 *
 */
static void run (const gchar *name,
  gint              nparams, 		// number of input params
  const GimpParam  *param, 			// input param array
  gint             *nreturn_vals, 	// pointer for setting the number of output params
  GimpParam       **return_vals){		// pointer to output param array

  // declare variables
  static GimpDrawable *drawable;
  static GimpParam  values[1];  // return variable
  GimpPDBStatusType status = GIMP_PDB_SUCCESS; // return status
  GimpRunMode       run_mode; // runnning mode of GIMP plugin
  gint x1 = 0;
  gint x2 = 0;
  gint y1 = 0;
  gint y2 = 0; // coordinates of the pixel regions
  gint channels = 0;

  // initialize logging
  initLogging();

  // setting the output parameters
  *nreturn_vals = 1;
  *return_vals  = values;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;

  /*  Get the specified drawable  */
  drawable = gimp_drawable_get (param[2].data.d_drawable);
  deblurImageData.drawable = drawable;

  /* Getting run_mode and determine how to retrieve the deblur parameters */
  run_mode = param[0].data.d_int32;

  // set initial values to defaults
  deblurParam = deblurParamDefaults;

  // get drawing coordinates
  gimp_drawable_mask_bounds (drawable->drawable_id, &x1, &y1, &x2, &y2);

  switch (run_mode){
    case GIMP_RUN_INTERACTIVE:
      /* Get options last values if needed */
      gimp_get_data ("Deblur-Pan-et-al.", &deblurParam);

      // retrieve the max kernel size from frameworkcontroller
      deblurParam.kernelMaxSize = getMaxKernelSize(x2-x1, y2-y1);
      if (deblurParam.kernelMaxSize > KERNELMAXSIZE ){
        (deblurParam.kernelMaxSize = KERNELMAXSIZE);
      }

      /* Display the dialog */
      if (! blur_dialog (&deblurParam, &deblurParamDefaults, drawable))
        return;
    break;

    case GIMP_RUN_NONINTERACTIVE:
      deblurImageData.runMode = nonInteractive;
      if (nparams != 11)
        status = GIMP_PDB_CALLING_ERROR;
      else
        // map parameters to the deblurParam struct
        mapInputParam(&deblurParam, param);
    break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Get options last values if needed  */
      gimp_get_data ("Deblur-Pan-et-al.", &deblurParam);
    break;

    default:
    break;
  }

  // initialize the progress bar
  gimp_progress_init ("Deblurring...");

  //Initiate deblur_results
  gimp_drawable_mask_bounds (drawable->drawable_id, &x1, &y1, &x2, &y2);
  initializePixelregions(x1, y1, x2 - x1, y2-y1);

  channels = gimp_drawable_bpp(drawable->drawable_id);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "channels is %d", channels);

  deblurImageData.x = x1;
  deblurImageData.y = y1;
  deblurImageData.imageHeight = y2-y1;
  deblurImageData.imageWidth = x2-x1;

  //initiate timer
  GTimer *timer = g_timer_new();
  logDeblurParam(deblurParam);

  // call deblur procedure in the adapter module to calculate the deblur
  gdouble res = deblurFromGimp(deblurParam, deblurImageData);

  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE,"res is %d", res);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE,"deblur() took %g seconds", g_timer_elapsed (timer, NULL));
  g_timer_destroy (timer);

  // if result need to be processed in Gimp
  if(res == 1){
	// process results in Gimp
    gimp_drawable_flush (drawable);
    gimp_drawable_merge_shadow ((drawable)->drawable_id, TRUE);
    gimp_drawable_update (drawable->drawable_id, x1,y1,x2-x1,y2-y1);
    gimp_displays_flush();
    gimp_drawable_detach(drawable);
  }

 /*  set used parameters in the core  */
 if (run_mode == GIMP_RUN_INTERACTIVE){
   gimp_set_data ("Deblur-Pan-et-al.", &deblurParam, sizeof (DeblurParam));
 }
 // close logging
 closeLogging();

 return;
}

/**
 * Creates the preview image based on the current settings entered in the deblurDialog.
 *
 * @param deblur_param 		pointer to the deblur parameter set in the deblur Dialog
 * @param GimpPreview		pointer to the preview drawable area used as output parameter for storing the preview results
 *
 */
void deblurPreview (DeblurParam *deblur_param, GimpPreview  *gimpPreview){
  // set runmode to preview
	deblurImageData.runMode = preview;

  // get preview position and size data
  gimp_preview_get_position (gimpPreview, &deblurImageData.x, &deblurImageData.y);
  gimp_preview_get_size (gimpPreview, &deblurImageData.imageWidth, &deblurImageData.imageHeight);

  // check if preview is not called without image
  if (deblurImageData.imageWidth>10){
	// create pixelregions for input and output
    initializePixelregions(deblurImageData.x, deblurImageData.y, deblurImageData.imageWidth,
							deblurImageData.imageHeight);
    // deblur the preview
    deblurFromGimp(deblurParam, deblurImageData);

    // update the preview region
    gimp_drawable_preview_draw_region (GIMP_DRAWABLE_PREVIEW (gimpPreview),
	                                          deblurImageData.latentPixelRgn);
  }

  // set runmode back to interactive
  deblurImageData.runMode = interactive;
}


/*
 * initializes the logging library
 */
static void initLogging (void){

  // setting up the log configuration
  mycat = log4c_category_get("deblur.pan");
  log4c_category_set_appender(mycat, log4c_appender_get("myrollingfileappender"));
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "logger initiated");
}

/*
 * Closes the logger and log file.
 */
static void closeLogging (void){
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "closing logger");
  log4c_fini();
}

/*
 * Logs the deblur parameters used for deblurring.
 *
 * @param deblurParam The DeblurParam struct containing the deblur parameter values.
 */
static void logDeblurParam(DeblurParam deblurParam){
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "kernelsize is %d", deblurParam.kernelSize);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "Gamma correction is %i", deblurParam.gammaCorrectionBoolean);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "Lambda pixel is %f", deblurParam.lambdaPixel);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "lambda gradient is %f", deblurParam.lambdaGrad);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "number of iterations is %d", deblurParam.numberIterations);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "lambda TV is %f", deblurParam.lambdaTV);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "lambda L0 is %f", deblurParam.lambdaL0);
  log4c_category_log(mycat, LOG4C_PRIORITY_NOTICE, "Weight ringing is %i", deblurParam.weightRing);
}

/*
 * Initializes the global pixelregions for the latent_image and blurred_image
 * based on the dimensions specified in the input parameters.
 *
 * @param x1	The x-coordinate of the upperleft corner of the pixelRegion.
 * @param y1	The y-coordinate of the upperleft corner of the pixelRegion.
 * @param width	The width of the pixelRegion.
 * @param height	The height of the pixelRegion.
 *
 */
static void initializePixelregions(gint x1, gint y1, gint width, gint height){
  // set latent image
  gimp_pixel_rgn_init (&latentPixelRgn,deblurImageData.drawable, x1, y1, width, height, TRUE, TRUE);

  // set blurred image
  gimp_pixel_rgn_init ( &blurredPixelRgn,deblurImageData.drawable,x1, y1, width, height,FALSE, FALSE);

  deblurImageData.latentPixelRgn = &latentPixelRgn;
  deblurImageData.blurredPixelRgn = &blurredPixelRgn;

}

/**
 * Copies the input array of Gimp containing the deblur parameter values
 * to the deblurParam struct.
 * @param deblurParam 	pointer to the deblurParam struct.
 * @param param			pointer to the GimpParam struct with the input parameters from GIMP.
 *
 */
static void mapInputParam(DeblurParam *deblurParam, GimpParam *param){
  deblurParam->kernelSize = param[3].data.d_int32;
  deblurParam->numberIterations = param[4].data.d_int32;
  deblurParam->lambdaGrad = param[5].data.d_float;
  deblurParam->lambdaPixel = param[6].data.d_float;
  deblurParam->lambdaTV = param[7].data.d_float;
  deblurParam->lambdaL0 = param[8].data.d_float;
  deblurParam->weightRing = param[9].data.d_int8;
  deblurParam->gammaCorrectionBoolean = param[10].data.d_int8;
}
