/*
 * ResultsDialog.c
 *
 * @author  T. Poppe
 * @author  E. van Schaijk
 * @version  1.0
 * @since  July, 2016
 *
 * Module which shows the deblur results as calculated by the domain layer.
 * The dialog shows the initial blurred image, the reconstructed kernel,
 * The intermediary result and the final reconstructed latent image.
 * All images are resized to max 200Px keeping their original width/height ratio
 */

  #include "ResultsDialog.h"

#include <libgimp/gimp.h>
  #include <libgimp/gimpui.h>
  #include <glib.h>
  #include <log4c.h>
  #include <gdk-pixbuf/gdk-pixbuf.h>

#include "DeblurAdapter.hpp"
#include "DeblurPanPlugin.h"

/**
 * Create a dialog which shows the calculated kernel, intermediate and latent image.
 * Based on the results in this dialog the user can press OK indicating that the result need to be
 * processed further or that Cancel indicating that the results show not be processed further
 *
 * @param deblurImageData A pointer to the struct holding the kernel, intermediary, latent and blurred images.
 *
 * @return boolean indicating the user action.
 * 					a one means it needs to be processed further,
 * 					a zero means the results don't need to be processed further.
 */
gboolean showResults(DeblurImageData deblurImageData){

  GtkWidget *dialog = NULL;
  GtkWidget *main_vbox = NULL,
		    *row1_hbox = NULL,
			*row2_hbox = NULL;
  GtkWidget *blurred_vbox = NULL,
		    *kernel_vbox = NULL,
			*intermediary_vbox = NULL,
			*latent_vbox = NULL;
  GtkWidget *kernel_image = NULL,
		    *intermediate_image = NULL,
			*blurred_image=NULL,
			*latent_image = NULL;
  GtkWidget *kernel_label = NULL,
		    *intermediate_label = NULL,
			*blurred_label = NULL,
			*latent_label = NULL;
  gboolean   run=0;

  gimp_ui_init ("deblurPanResults", FALSE);

  dialog = gimp_dialog_new ("Deblur Results", "deblurResults",
		  NULL, 0,NULL, NULL, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);

  // create the main box holding all gui components
  main_vbox = gtk_vbox_new (FALSE, 6);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), main_vbox);
  gtk_widget_show (main_vbox);

  // create horizontal box for first two images
  row1_hbox = gtk_hbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (row1_hbox), 12);
  gtk_widget_show (row1_hbox);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), row1_hbox);

  // create vertical box for blurred image
  blurred_vbox = gtk_vbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (blurred_vbox), 12);
  gtk_widget_show (blurred_vbox);
  gtk_container_add (GTK_CONTAINER (row1_hbox), blurred_vbox);

  // create label blurred_label
  blurred_label = gtk_label_new_with_mnemonic ("Blurred Image");
  gtk_box_pack_start (GTK_BOX (blurred_vbox), blurred_label, FALSE, FALSE, 6);
  gtk_label_set_justify (GTK_LABEL (blurred_label), GTK_JUSTIFY_RIGHT);
  gtk_widget_show (blurred_label);

  // create blurred image
  blurred_image = gtk_image_new_from_pixbuf (deblurImageData.blurredPixbuf);
  gtk_box_pack_start (GTK_BOX (blurred_vbox), blurred_image, TRUE, TRUE, 0);
  gtk_widget_show (blurred_image);

  // create vertical box for kernel image
  kernel_vbox = gtk_vbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (kernel_vbox), 12);
  gtk_widget_show (kernel_vbox);
  gtk_container_add (GTK_CONTAINER (row1_hbox), kernel_vbox);


  // create label kernel
  kernel_label = gtk_label_new_with_mnemonic ("Kernel image");
  gtk_box_pack_start (GTK_BOX (kernel_vbox), kernel_label, FALSE, FALSE, 6);
  gtk_label_set_justify (GTK_LABEL (kernel_label), GTK_JUSTIFY_RIGHT);
  gtk_widget_show (kernel_label);

  // create kernel image
  kernel_image = gtk_image_new_from_pixbuf (deblurImageData.kernelPixbuf);
  gtk_box_pack_start (GTK_BOX (kernel_vbox), kernel_image, TRUE, TRUE, 0);
  gtk_widget_show (kernel_image);

  // create horizontal box for second row images
  row2_hbox = gtk_hbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (row2_hbox), 12);
  gtk_widget_show (row2_hbox);
  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), row2_hbox);

  // create vertical box for intermediary image
  intermediary_vbox = gtk_vbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (intermediary_vbox), 12);
  gtk_widget_show (intermediary_vbox);
  gtk_container_add (GTK_CONTAINER (row2_hbox), intermediary_vbox);

  // create label intermediate
  intermediate_label = gtk_label_new_with_mnemonic ("Intermediate Latent Image");
  gtk_box_pack_start (GTK_BOX (intermediary_vbox), intermediate_label, FALSE, FALSE, 6);
  gtk_label_set_justify (GTK_LABEL (intermediate_label), GTK_JUSTIFY_RIGHT);
  gtk_widget_show (intermediate_label);

  // create kernel image
  intermediate_image = gtk_image_new_from_pixbuf (deblurImageData.intermediaryPixbuf);
  gtk_box_pack_start (GTK_BOX (intermediary_vbox), intermediate_image, TRUE, TRUE, 0);
  gtk_widget_show (intermediate_image);

  // create vertical box for latent image
  latent_vbox = gtk_vbox_new (FALSE, 12);
  gtk_container_set_border_width (GTK_CONTAINER (latent_vbox), 12);
  gtk_widget_show (latent_vbox);
  gtk_container_add (GTK_CONTAINER (row2_hbox), latent_vbox);

  // create label latent_label
  latent_label = gtk_label_new_with_mnemonic ("Latent Image");
  gtk_box_pack_start (GTK_BOX (latent_vbox), latent_label, FALSE, FALSE, 6);
  gtk_label_set_justify (GTK_LABEL (latent_label), GTK_JUSTIFY_RIGHT);
  gtk_widget_show (latent_label);

  // create latent image
  latent_image = gtk_image_new_from_pixbuf (deblurImageData.latentPixbuf);
  gtk_box_pack_start (GTK_BOX (latent_vbox), latent_image, TRUE, TRUE, 0);
  gtk_widget_show (latent_image);


  gtk_widget_show (dialog);

  run = (gimp_dialog_run (GIMP_DIALOG (dialog)) == GTK_RESPONSE_OK);

  gtk_widget_destroy (dialog);

  return run;
}
